import copy
import requests
from typing import Union

from ..._common.config import *
from ..._common.const import UniverseFrequencyType as _UniverseFrequencyType
from ._list import _list_dataitem
from .preference import get_quiet_mode
from ..._utils import (
    _validate_args,
    get,
    post,
    patch,
    delete,
    get_sm_attributevalue,
    _authentication,
    _process_response,
    _process_fileresponse,
    _fetch_and_parse,
    plot_tree,
    are_periods_exclusive,
)
from ..._utils.exceptions import PrismNotFoundError, PrismValueError


__all__ = [
    "list_universe",
    "get_universe",
    "upload_universe",
    "save_index_as_universe",
    "combine_universe",
    "rename_universe",
    "delete_universe",
    "get_universe_template",
    "filter_universe",
]


@_validate_args
def list_universe(search: str = None, tree=False):
    """
    Return the all usable universes.

    Parameters
    ----------
        search : str, default None
            | Search word for universe name, the search is case-insensitive.
            | If None, the entire list is returned.

        tree : bool, default False
            | If True, the folder structure of the universes is visualized in a UNIX-style.

    Returns
    -------
        str or pandas.DataFrame
            | If tree is True, print file system tree of universes.
            | If tree is False, return usable universes in DataFrame.
            | Columns:

            - *universeid*
            - *universename*
            - *universetype*
            - *startdate*
            - *enddate*

    Examples
    --------
        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31
    """
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv", search)
    if universes_df.empty:
        raise PrismNotFoundError("No Universes found.")
    if not tree:
        return universes_df
    plot_tree(universes_df["universename"].tolist(), "universe")


@_validate_args
def get_universe(
    universe,
    startdate: str = None,
    enddate: str = None,
    frequency: str = None,
    shownid: list = None,
):
    """
    Return the specified universe in dataframe.

    Parameters
    ----------
        universe : str or int
            Universe name (str) or universe id (int) to get.

        startdate : str, default None
            | Start date of the data. The data includes start date.
            | If None, the start date of the universe is used. The same applies when the input is earlier than the universe start date.

        enddate : str, default None
            | End date of the data. The data excludes end date.
            | If None, the end date of the universe is used. The same applies when the input is later than the universe end date.

        frequency : str, {'D', 'W', 'MS', 'SMS', 'SM', 'M', 'Q', 'QS', 'AS', 'A'}
            | Desired sampling frequency of universe constituents. If expand is False, this parameter is ignored.

        shownid : list, default None
            | List of Security Master attributes to display with the data.
            | If None, default attributes set in preferences is shown.
            | If empty list ([]), no attribute is shown.

    Returns
    -------
        pandas.DataFrame
            Universe.
            Columns:
            - *listingid*
            - *date*
            - *attributes provided by the user via 'shownid'*

    Examples
    --------
        >>> prism.list_universe()
           universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

        >>> prism.get_universe(univese='Russell 3000 Index')
                  listingid        date
        0           2585895  1978-12-29
        1           2586016  1978-12-29
        2           2586064  1978-12-29
        3           2586086  1978-12-29
        4           2586118  1978-12-29
        ...             ...         ...
        10110503  701835357  2199-12-31
        10110504  701932931  2199-12-31
        10110505  703822433  2199-12-31
        10110506  704721046  2199-12-31
        10110507  706171023  2199-12-31

        >>> prism.get_universe('Russell 3000 Index', startdate='2010-01-01', enddate='2015-12-31', shownid=['companyname'])
                listingid        date                   companyname
        0          2585893  2010-01-03                      AAON INC
        1          2585895  2010-01-03                      AAR CORP
        2          2585957  2010-01-03    ADC TELECOMMUNICATIONS INC
        3          2586016  2010-01-03            ABM INDUSTRIES INC
        4          2586068  2010-01-03            AEP INDUSTRIES INC
        ...            ...         ...                           ...
        2194810  325621650  2015-12-27        AVAGO TECHNOLOGIES LTD
        2194811  325832671  2015-12-27                     POZEN INC
        2194812  326004249  2015-12-27  LIBERTY INTERACTV CP QVC GRP
        2194813  344286611  2015-12-27            ITT INDUSTRIES INC
        2194814  365743684  2015-12-27     HERTZ GLOBAL HOLDINGS INC
    """

    universeid, _ = parse_universe_to_universeid(universe)

    universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
    universe_startdate = universe_info["Start Date"].values[0]
    universe_enddate = universe_info["End Date"].values[0]

    universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

    if universe_period_violated:
        raise PrismValueError(
            f'Universe query period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
        )

    if (frequency is not None) and (startdate is None):
        raise PrismValueError(f"Please provide startdate to resample with frequency {frequency}")

    default_shownid = True
    if (shownid is not None) and (len(shownid) == 0):
        shownid = None
        default_shownid = False
    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]

    params = {
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "default_shownid": default_shownid,
        "frequency": frequency,
    }
    headers = _authentication()
    res = requests.get(URL_UNIVERSES + f"/{universeid}", params=params, headers=headers)
    ret, _ = _process_fileresponse(res, res.content)
    return list(ret.values())[0]


@_validate_args
def upload_universe(universefile: str, universename: str):
    """
    Create a new universe with a csv file specifying the universe constituents. The csv file should have four columns: valuetype, value, startdate, enddate.
    “valuetype” specifies the Security Master attribute. “value” indicates the value of such attribute to include into the universe. “startdate” and “enddate” indicates the time range of specified security to be included into the universe.

    Parameters
    ----------
        universefile: str
            Path to the csv file to upload.
        universename: str
            Name of the universe to be created.

    Returns
    -------
        status: dict
            Status of upload_universe.

    Examples
    --------
        >>> import pandas as pd
        >>> universe_df = pd.read_csv("./my_universe.csv")
        >>> universe_df # include all securities being traded in the Korean Exchange
        valuetype  value   startdate     enddate
        0        MIC   XKRX  1700-01-01  2999-12-31

        >>> prism.upload_universe(universefile="./my_universe.csv", universename="krx")
        {'status': 'Success',
        'message': 'Universe saved',
        'result': [{'resulttype': 'universeid', 'resultvalue': 1}]}

        >>> prism.list_universe()
        universeid  universename  universetype   startdate     enddate
        0           1           krx        custom  1700-01-01  2199-12-31

    """
    should_overwrite, err_msg = should_overwrite_universe(universename, "uploading")
    if not should_overwrite:
        return err_msg
    universefile_params = {"universefile": open(universefile, "rb")}
    headers = {"Authorization": _authentication()["Authorization"], "client": "python"}
    res = requests.post(
        url=URL_UNIVERSES + "/flatfile",
        params={"path": universename + ".puv"},
        files=universefile_params,
        headers=headers,
    )
    ret = _process_response(res)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def save_index_as_universe(
    dataitemid: int,
    startdate: str = None,
    enddate: str = None,
    universename: str = None,
):
    """
    Create a new universe containing the constituents of an index.

    Parameters
    ----------
        dataitemid : int
            Unique identifier for the different data item. This identifies the index.
        startdate: str, default None
            Start date of the data. The data includes start date.
            If None, the start date of the universe is used. The same applies when the input is earlier than the universe start date.
        enddate: str, default None
            End date of the data. The data excludes end date.
            If None, the end date of the universe is used. The same applies when the input is later than the universe end date.
        universename: str, default None
            Name of the universe to be saved from the index.
            If None, the index name is used.


    Returns
    -------
        status: dict
            Status of save_index_as_universe.

    Examples
    --------
        >>> prism.index.universe_dataitems("S&P 500")
            dataitemid                                       dataitemname    datamodule       package
        0       4006682                                            S&P 500  S&P US Index  S&P US Index
        1       4006683           S&P 500 - Alternative Carriers (Sub Ind)	S&P US Index  S&P US Index
        2       4006684                 S&P 500 - Biotechnology (Industry)	S&P US Index  S&P US Index
        3       4006685                  S&P 500 - Biotechnology (Sub Ind)	S&P US Index  S&P US Index
        4       4006686                   S&P 500 - Broadcasting (Sub Ind)  S&P US Index  S&P US Index
        ...	        ...	                                               ...           ...           ...
        308     4006990                  S&P 500 Water Utilities (Sub Ind)	S&P US Index  S&P US Index
        309     4006991  S&P 500 Wireless Telecommunication Services (I...  S&P US Index  S&P US Index
        310     4006992  S&P 500 Wireless Telecommunication Services (S...  S&P US Index  S&P US Index
        311     4006993                      S&P 500 Oil (Composite) Index	S&P US Index  S&P US Index
        312     4006994  S&P 500 Semiconductors (Sub Ind)(30-APR-03) In...  S&P US Index  S&P US Index

        >>> prism.save_index_as_universe(dataitemid=4006682)
        {'status': 'Success',
        'message': 'Universe saved',
        'result': [{'resulttype': 'universeid', 'resultvalue': 1}]}

        >>> prism.list_universe()
        universeid  universename  universetype   startdate     enddate
        0           1       S&P 500         index  1700-01-01  2199-12-31
    """

    if universename is None:
        indices = _list_dataitem(
            datacategory = "Index",
            datacomponent = "Index Level",
        )
        universename = indices[indices["dataitemid"] == dataitemid]["dataitemname"].values[0]

    should_overwrite, err_msg = should_overwrite_universe(universename, "combining")
    if not should_overwrite:
        return err_msg
    params = {
        "dataitemid": dataitemid,
        "startdate": startdate,
        "enddate": enddate,
        "path": universename + ".puv",
    }
    ret = post(URL_UNIVERSES + f"/index", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def combine_universe(universes: list, newuniversename: str):
    """
    Create a new universe by combining existing universes.

    Parameters
    ----------
        universes: list of int or list of str
            List of universe id (int) or universe name (str) to combine.
        newuniversename: str
            Name of the universe to be created.

    Returns
    -------
        status: dict
            Status of combine universe.

    Examples
    --------
        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

        >>> prism.combine_universe(["Korea Stock Price 200 Index", "S&P 500"], newuniversename="kospi_snp")
        {'status': 'Success',
        'message': 'Universe saved',
        'result': [{'resulttype': 'universeid', 'resultvalue': 1}]}

        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31
        4           5                    kospi_snp        custom  1700-01-01  2199-12-31

    """
    if len(universes) < 2:
        SystemExit("Please provide more than 2 universes to combine.")
    should_overwrite, err_msg = should_overwrite_universe(newuniversename, "saving")
    if not should_overwrite:
        return err_msg
    universeids = parse_universes_to_universeids(universes)
    params = {"universeids": universeids, "path": newuniversename + ".puv"}
    ret = post(URL_UNIVERSES + f"/combine", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_universe(old: Union[str, int], new: str):
    """
    Rename universe. Location of universe within the folder structure can be changed using the method.

    Parameters
    ----------
        old: str or int
            Name of existing universe (str) or universe id (int) to be renamed.
        new: str
            New name of the universe.

    Returns
    -------
        status: dict
            Status of rename universe.

    Examples
    --------
        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

        >>> prism.rename_universe(old='S&P 500', new='newname_snp')
        {'status': 'Success',
        'message': 'Universe renamed',
        'result': [{'resulttype': 'universeid', 'resultvalue': 2}]}

        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                  newname_snp         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

    """
    should_overwrite, err_msg = should_overwrite_universe(new, "renaming")
    if not should_overwrite:
        return err_msg
    universeid, _ = parse_universe_to_universeid(old)
    ret = patch(URL_UNIVERSES + f"/{universeid}", {"newpath": new + ".puv"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_universe(universe: Union[str, int]):
    """
    Delete universe.

    Parameters
    ----------
        universe : str or int
            Universe name (str) or universe id (int) to delete.

    Returns
    -------
        status: dict
            Status of delete_universe.

    Examples
    --------
        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
        3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

        >>> prism.delete_universe(universe='Russell 3000 Growth Index')
        {'status': 'Success',
        'message': 'Universe deleted',
        'result': [{'resulttype': 'universeid', 'resultvalue': 3}]}

        >>> prism.list_universe()
        universeid                 universename  universetype   startdate     enddate
        0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
        1           2                      S&P 500         index  1700-01-01  2199-12-31
        2           4           Russell 3000 Index         index  1700-01-01  2199-12-31
    """
    universeid, _ = parse_universe_to_universeid(universe)
    ret = delete(URL_UNIVERSES + f"/{universeid}")
    print(ret["message"])
    return ret


def get_universe_template():
    """
    Get list of columns required for prism.upload_universe.

    Returns
    -------
        columns : list of str
            List of columns required for csv file to be uploaded.

    Examples
    --------
        >>> prism.get_universe_template()
        valuetype, value, startdate, enddate
    """
    res = requests.get(url=URL_UNIVERSES + "/template", headers=_authentication())
    return res.text


@_validate_args
def filter_universe(condition: list, universename: str):
    """
    Create a new universe from the entire Security Master using simple filter rules applied on Security Master attribute.

    Parameters
    ----------
        condition : list of dict
            List of security conditions to apply as filter. Each dictionary contains three keys: “operator”, “attribute”, “value”. Value must be a list

            - operator: {”AND”, “OR”}
            - attribute: Security Master attribute to condition on.
            - value: Value of Security Master attribute to condition on.
            - not: bool

        universename : str
            Name of the universe to be created.

    Returns
    -------
        status: dict
            Status of filter_universe.

    Examples
    --------
        >>> condition = [
            {'attribute': 'country', 'value': ['AU', 'NZ', 'HK', 'SG', 'JP', 'IN', 'ID', 'KR', 'MY', 'PH', 'TW', 'TH']},
            {'attribute': 'CIQ primary', 'value': ['primary'], 'operator': 'AND'}
            ]
        >>> prism.filter_universe(condition=condition, universename="APAC_primary")
        {'status': 'Success',
        'message': 'Universe saved',
        'result': [{'resulttype': 'universeid', 'resultvalue': 1}]}

        >>> prism.list_universe()
        universeid  universename  universetype   startdate     enddate
        0           1  APAC_primary         index  1700-01-01  2199-12-31
    """
    if len(condition) == 0:
        raise PrismValueError("At least one condition must be provided.")
    should_overwrite, err_msg = should_overwrite_universe(universename, "filtering")
    if not should_overwrite:
        return err_msg
    condition_copy = copy.deepcopy(condition)
    for i in range(len(condition)):
        assert isinstance(condition[i], dict), "condition must be list of dicts"
        assert set(condition[i].keys()) - {"operator", "isnot"} == {
            "attribute",
            "value"
        }, 'Valid arguments are "attribute", "value", "operator", and "isnot"'
        if not isinstance(condition[i]["value"], list):
            raise PrismValueError("Value must be a list")
        condition_copy[i]["attribute"] = get_sm_attributevalue(condition_copy[i]["attribute"])
        condition_copy[i]["isnot"] = condition[i].get("isnot", False)
    ret = post(URL_UNIVERSES + "/filter", {"path": universename + ".puv"}, condition_copy)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def parse_universe_to_universeid(universe: Union[str, int]):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        raise PrismNotFoundError("No universe found.")

    if isinstance(universe, int):
        universes_df = universes_df[universes_df["universeid"] == universe]
    elif isinstance(universe, str):
        universes_df = universes_df[universes_df["universename"] == universe]
    else:
        raise PrismValueError("Please provide universe name or universe id for universe.")

    if universes_df.empty:
        raise PrismNotFoundError(f"No universe matching: {universe}", True, None)
    universeid = universes_df["universeid"].values[0]
    universename = universes_df["universename"].values[0]

    return universeid, universename


def parse_universes_to_universeids(universes: list):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        raise PrismNotFoundError("No universe found.")

    universenames = []
    universeids = []
    for u in universes:
        if isinstance(u, int):
            universeids.append(u)
        elif isinstance(u, str):
            universenames.append(u)
        else:
            raise PrismValueError("Please provide universe name or universe id for universe.")

    ret_universeids = set()
    if len(universenames) != 0:
        matching_universes_df = universes_df[universes_df["universename"].isin(universenames)]
        if len(matching_universes_df) != len(universenames):
            not_matched = list(set(universenames) - set(matching_universes_df["universename"].tolist()))
            raise PrismNotFoundError(f"No universe matching: {not_matched}")
        universeids_from_universenames = matching_universes_df["universeid"].tolist()
        ret_universeids.update(universeids_from_universenames)

    if len(universeids) != 0:
        matching_universes_df = universes_df[universes_df["universeid"].isin(universeids)]
        if len(matching_universes_df) != len(universeids):
            not_matched = list(set(universeids) - set(matching_universes_df["universeid"].tolist()))
            raise PrismNotFoundError(f"No universe matching: {not_matched}")
        existing_universeids = matching_universes_df["universeid"].tolist()
        ret_universeids.update(existing_universeids)
    return list(ret_universeids)


def should_overwrite_universe(universename, operation):
    if get_quiet_mode():
        return True, None
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        return True, None

    if universename in universes_df["universename"].to_list():
        overwrite = input(f"{universename} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Not {operation} universe."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None



from ..._common.config import *
from ..._common.const import *
from ..._utils import _validate_args, get, delete, post


__all__ = ['list_job', 'get_job', 'delete_job', 'strategy_backtest_jobs', 'factor_backtest_jobs', 'screen_jobs', 'extract_job', 'cancel_job', 'prioritise_job', 'export_data_jobs']


@_validate_args
def list_job(
    tasktype: str = None,
    jobstatus: str = None,
    startrange: str = None,
    endrange: str = None,
):
    """
    Return all jobs.

    Parameters
    ----------
    tasktype : str, {”screen”, “factor_backtest”}
        | Task type is a value from the list.

    jobstatus : str, {“Completed”, “Failed”}
        | Job status is a value from the list.

    startrange : str, default None
        | Date string to query list of jobs within the date range.

    endrage : str, default None
        | Date string to query list of jobs within the date range.

    Returns
    -------
        pandas.DataFrame
            return all queried jobs.
        Columns:
            - *jobid*
            - *jobname*
            - *jobstatus*
            - *starttime*
            - *endtime*
            - *tasktype*
            - *result*
            - *description*

    Examples
    --------
    >>> prism.list_job()
       jobid            jobname  jobstatus                starttime                  endtime         tasktype                                        result  description
    0      6  factor_backtest_6  Completed  2022-08-10 17:38:57.290  2022-08-10 17:44:24.599  factor_backtest  {'factor_backtest': {'factorbacktestid': 6}}         None
    1      5  factor_backtest_5  Completed  2022-07-31 14:28:43.679  2022-07-31 14:29:18.272  factor_backtest  {'factor_backtest': {'factorbacktestid': 5}}         None
    2      4  factor_backtest_4  Completed  2022-07-29 15:15:49.582  2022-07-29 15:16:24.828  factor_backtest  {'factor_backtest': {'factorbacktestid': 4}}         None
    3      3  factor_backtest_3  Completed  2022-07-25 09:26:24.777  2022-07-25 09:27:01.571  factor_backtest  {'factor_backtest': {'factorbacktestid': 3}}         None
    4      2  factor_backtest_2  Completed  2022-07-18 04:47:02.414  2022-07-18 04:47:43.182  factor_backtest  {'factor_backtest': {'factorbacktestid': 2}}         None
    5      1  factor_backtest_1  Completed  2022-07-05 17:19:41.197  2022-07-05 17:19:53.012  factor_backtest  {'factor_backtest': {'factorbacktestid': 1}}         None

    """
    param = {'jobstatus': jobstatus, 'tasktype': tasktype, 'startrange': startrange, 'endrange': endrange}
    return get(URL_JOBS, param)


@_validate_args
def get_job(jobid: int):
    """
    Get detail of specified job.

    Parameters
    ----------
        jobid : int
            Specify the job id.

    Returns
    -------
        data : dataframe
            | Return job detail. The columns depends on the task type of the job
            | If tasktype is factor backtest

            - *jobid*
            - *jobname*
            - *jobstatus*
            - *starttime*
            - *endtime*
            - *universeid*
            - *universepath*
            - *factorbacktestid*
            - *avg_turnover*
            - *avg_ic*
            - *top_bottom_spread*
            - *frequency*
            - *bins*
            - *max_days*
            - *rank_method*
            - *description*
            - *period*

            | If tasktype is screen

            - *jobid*
            - *jobname*
            - *jobstatus*
            - *starttime*
            - *endtime*
            - *frequency*
            - *screeneduniverseid*
            - *screeneduniversepath*
            - *universepath*
            - *universeid*
            - *description*
            - *period*

    Examples
    --------
    >>> prism.list_job()
       jobid            jobname  jobstatus                starttime                  endtime         tasktype                                        result  memo
    0      6  factor_backtest_6  Completed  2022-08-10 17:38:57.290  2022-08-10 17:44:24.599  factor_backtest  {'factor_backtest': {'factorbacktestid': 6}}  None
    1      5  factor_backtest_5  Completed  2022-07-31 14:28:43.679  2022-07-31 14:29:18.272  factor_backtest  {'factor_backtest': {'factorbacktestid': 5}}  None
    2      4  factor_backtest_4  Completed  2022-07-29 15:15:49.582  2022-07-29 15:16:24.828  factor_backtest  {'factor_backtest': {'factorbacktestid': 4}}  None
    3      3  factor_backtest_3  Completed  2022-07-25 09:26:24.777  2022-07-25 09:27:01.571  factor_backtest  {'factor_backtest': {'factorbacktestid': 3}}  None
    4      2  factor_backtest_2  Completed  2022-07-18 04:47:02.414  2022-07-18 04:47:43.182  factor_backtest  {'factor_backtest': {'factorbacktestid': 2}}  None
    5      1  factor_backtest_1  Completed  2022-07-05 17:19:41.197  2022-07-05 17:19:53.012  factor_backtest  {'factor_backtest': {'factorbacktestid': 1}}  None

    >>> prism.get_job(1)
         jobid            jobname  jobstatus  factorbacktestid  avg_turnover  avg_ic  top_bottom_spread  frequency  bins  max_days  rank_method  description                   period
    0	     1  factor_backtest_1  Completed                31      0.475282  0.0178          -0.000383          Q    10        93     standard         None  2013-01-01 ~ 2015-01-01

    """
    return get(f'{URL_JOBS}/id/{jobid}', None)


@_validate_args
def strategy_backtest_jobs():
    return get(URL_JOBS + '/type/strategy_backtest', None)


@_validate_args
def delete_job(jobid: int):
    """
    Delete job with job id.

    Parameters
    ----------
        jobid : int
            Specify job id to be deleted.

    Returns
    -------
        status : dict
            Status of delete_job.

    Examples
    --------
    >>> prism.list_job()
       jobid            jobname  jobstatus                starttime                  endtime         tasktype                                        result  description
    0      6  factor_backtest_6  Completed  2022-08-10 17:38:57.290  2022-08-10 17:44:24.599  factor_backtest  {'factor_backtest': {'factorbacktestid': 6}}         None
    1      5  factor_backtest_5  Completed  2022-07-31 14:28:43.679  2022-07-31 14:29:18.272  factor_backtest  {'factor_backtest': {'factorbacktestid': 5}}         None
    2      4  factor_backtest_4  Completed  2022-07-29 15:15:49.582  2022-07-29 15:16:24.828  factor_backtest  {'factor_backtest': {'factorbacktestid': 4}}         None
    3      3  factor_backtest_3  Completed  2022-07-25 09:26:24.777  2022-07-25 09:27:01.571  factor_backtest  {'factor_backtest': {'factorbacktestid': 3}}         None
    4      2  factor_backtest_2  Completed  2022-07-18 04:47:02.414  2022-07-18 04:47:43.182  factor_backtest  {'factor_backtest': {'factorbacktestid': 2}}         None
    5      1  factor_backtest_1  Completed  2022-07-05 17:19:41.197  2022-07-05 17:19:53.012  factor_backtest  {'factor_backtest': {'factorbacktestid': 1}}         None

    >>> prism.delete_job(1)
    {
    'status': 'Success',
    'message': 'Job deleted',
    'result': [{'resulttype': 'jobid', 'resultvalue': 1}]
    }

    >>> prism.list_job()
       jobid            jobname  jobstatus                starttime                  endtime         tasktype                                        result  description
    0      6  factor_backtest_6  Completed  2022-08-10 17:38:57.290  2022-08-10 17:44:24.599  factor_backtest  {'factor_backtest': {'factorbacktestid': 6}}         None
    1      5  factor_backtest_5  Completed  2022-07-31 14:28:43.679  2022-07-31 14:29:18.272  factor_backtest  {'factor_backtest': {'factorbacktestid': 5}}         None
    2      4  factor_backtest_4  Completed  2022-07-29 15:15:49.582  2022-07-29 15:16:24.828  factor_backtest  {'factor_backtest': {'factorbacktestid': 4}}         None
    3      3  factor_backtest_3  Completed  2022-07-25 09:26:24.777  2022-07-25 09:27:01.571  factor_backtest  {'factor_backtest': {'factorbacktestid': 3}}         None
    4      2  factor_backtest_2  Completed  2022-07-18 04:47:02.414  2022-07-18 04:47:43.182  factor_backtest  {'factor_backtest': {'factorbacktestid': 2}}         None

    """
    ret = delete(URL_JOBS + f'/{jobid}')
    return ret


@_validate_args
def cancel_job(jobid: int):
    """
    Cancel specified job.

    Parameters
    ----------
        jobid : int
            Specify the job id.

    Returns
    -------
        Result of canceled job.

    Examples
    --------
    >>> ep = prism.financial.income_statement(100723, 'Q').resample(frequency='M', lookback=30) / prism.market.market_cap().resample('M', lookback=30)
    >>> prism.factor_backtest(ep, 'Korea_primary', 'M', 10, '2015-01-01').run('ep_df')
    factor_backtest is added to worker queue!: jobid is 321

    >>> prism.cancel_job(321)
    {
        'status': 'Canceled',
        'message': 'screen is deleted from queue!',
        'result': []
    }
    """
    ret = delete(URL_TASK + f'/revoke/{jobid}')
    return ret


@_validate_args
def prioritise_job(jobid: int):
    """
    Prioritise specified job.

    Parameters
    ----------
        jobid : int
            Specify the job id.

    Returns
    -------
        Result of prioritized job.


    Examples
    --------
    >>> ep = prism.financial.income_statement(100723, 'Q').resample(frequency='M', lookback=30) / prism.market.market_cap().resample('M', lookback=30)
    >>> prism.factor_backtest(ep, 'Korea_primary', 'M', 10, '2015-01-01').run('ep_df')
    factor_backtest is added to worker queue!: jobid is 321

    >>> prism.prioritise_job(321)
    {
        'status': 'Success',
        'message': 'factor backtest is prioritised in the queue!',
        'result': []
    }
    """
    ret = post(URL_TASK + f'/prioritise/{jobid}')
    return ret


@_validate_args
def extract_job(jobid: int, return_code=False):
    """
    Generate code which reproduces the taskquery which runs the job specified by the jobid. If return_code is False, the method returns None and prints the code.

    Parameters
    ----------
        jobid : int
            | Job id whose taskquery is to be extracted.

        return_code : bool, default False
            | if True, the method returns the code. Else the code is printed as system output, returning None.

    Returns
    -------
        return_code = True
            code: str

        return_code = False
            None

    Examples
    --------
    >>> ni = prism.financial.income_statement(100639, periodtype="LTM")
    >>> mcap = prism.market.market_cap()
    >>> ep = ni / mcap
    >>> ep_fb = prism.factor_backtest(factor=ep, universe="S&P 500", startdate="2013-01-01", enddate="2015-01-01", frequency="Q", bins=10)
    >>> ep_fb.run()
    Done!
    factor_backtest Completed: factorbacktestid is 1
    Fetching A Link to Factor Backtest Report...
    Link to Factor Backtest Report:
    https://ext.prism39.com/report/factor_backtest/my_username_1_96cebcc1-fe3d-4de7-b433-e32d6fcc13c9/

    >>> prism.extract_job(jobid=1)
    x0 = prism.financial.income_statement(dataitemid=100639, periodtype="LTM", package="CIQ Premium Financials", preliminary=True, currency=None)
    x1 = prism.market.marketcap(currency=None, package="CIQ Market")
    x2 = x0 / x1
    x3 = prism.factor_backtest(factor=x2, universe=2, startdate="2013-01-01", enddate="2015-01-01", frequency="Q", bins=10, rank_method="standard", max_days=None)
    """
    code = get(URL_JOBS + f'/extract/{jobid}', {'dialect': 'python'})
    if return_code:
        return code
    else:
        print(code)


@_validate_args
def rename_job(jobid: int, jobname: str):
    return post(URL_JOBS + f'/jobname/{jobid}', body=jobname)


@_validate_args
def add_job_description(jobid: int, description: str):
    return post(URL_JOBS + f'/description/{jobid}', body=description)



import os
from typing import Literal

from .._common.config import StateSettings, ROOT_EXT_WEB_URL, URL


class Version():
    # _regex = ''
    def __init__(self) -> None:
        self._version, self._environment = self.load_setup()
        self._api_url, self._gui_url = self.load_url()

    def __repr__(self) -> str:
        return f"<Version {self.version}-{self.environment}>"

    def __str__(self) -> str:
        return "\n".join(
            [
                f"Version: {self.version}",
                f"Environment: {self.environment}",
                f"API_URL: {self.api_url}",
                f"GUI_URL: {self.gui_url}",
            ]
        )

    @property
    def version(self) -> str:
        return self._version

    @property
    def environment(self) -> Literal["production", "dev", "stg"]:
        return self._environment

    @property
    def api_url(self) -> str:
        return self._api_url

    @property
    def gui_url(self) -> str:
        return self._gui_url

    def load_setup(self)-> tuple[Literal, Literal["production", "dev", "stg"]]:
        dotenv_settings = StateSettings()
        env = os.environ.get('ENV_STATE', dotenv_settings.ENV_STATE)
        ver = dotenv_settings.VERSION
        if env == "prod":
            env = 'production'

        return ver, env

    def load_url(self):
        api = URL
        gui = ROOT_EXT_WEB_URL
        return api, gui


__version__ = str(Version())


from datetime import date

import pytest

from prism._common.const import DateType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def _build_event(
    datacomponentclass,
    dataitemid: int,
    datetype: str,
    package: str = None,
) -> DateType:
    return datetype


def test_datetype_param():
    assert _build_event(None, 1, datetype=None) is None
    assert _build_event(None, 1, "announceddate",
                        "Prism Market") == DateType.ANNOUNCEDDATE

    # check type
    assert isinstance(_build_event(None, 1, "announceddate", "Prism Market"), DateType)

    # valid datetype param
    assert _build_event("news", 1, "entereddate") == DateType.ENTEREDDATE
    assert _build_event("news", 2, "announceddate") == DateType.ANNOUNCEDDATE

    assert _build_event(
        "news", 3, datetype="entereddate") == DateType.ENTEREDDATE
    assert _build_event(
        "news", 4, datetype="announceddate") == DateType.ANNOUNCEDDATE


def test_wrong_datetype_param():
    # wrong value
    with pytest.raises(PrismValueError):
        assert _build_event("news", 1, datetype="")

    with pytest.raises(PrismValueError):
        assert _build_event("news", 1, datetype="1998/10/03")

    with pytest.raises(PrismValueError):
        assert _build_event("news", 1, datetype="2023.04.04")

    with pytest.raises(PrismValueError):
        assert _build_event(
            "news", 23, datetype="ENTEREDDATE") == DateType.ENTEREDDATE

    with pytest.raises(PrismValueError):
        assert _build_event(
            "news", 23, datetype="AnnouncedDate") == DateType.ANNOUNCEDDATE

    # wrong type
    with pytest.raises(PrismTypeError):
        assert _build_event("news", "123", datetype=None)

    with pytest.raises(PrismTypeError):
        assert _build_event("news", 1, datetype=20220101)

    with pytest.raises(PrismTypeError):
        assert _build_event("news", 1, datetype=date.today())



from .._req_builder import _list_dataitem
from ..._common.const import (
    CompanyRelAttributeType as _CompanyRelAttributeType,
    CurrencyTypeWithReportTrade as _CurrencyTypeWithReportTrade
)
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils import _get_params, _validate_args
from ..._utils.exceptions import PrismValueError


__all__ = [
    "ma",
    "buyback",
    "dataitems",
]


_data_category = __name__.split(".")[-1]


class _PrismTransactionComponent(_PrismDataComponent, _PrismComponent):
    _component_category_repr = _data_category

    @classmethod
    def _dataitems(cls, search : str = None, package : str = None):
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )


class ma(_PrismTransactionComponent):
    """
    | Get merger and acquisition transaction-related data for equity securities.
    | The default frequency is aperiodic.

    Parameters
    ----------
    dataitemid : int
        Unique identifier for the specific data item.

    target : bool, default True
        Specifies whether the company is a target for an acquisition or an acquirer:
        - If True, the data will be presented from the perspective of a target for an acquisition. This means that given a universe, the data will display all companies that made bids to the companies within the universe.
        - If False, the data will be presented from the perspective of a acquirer for an acquisition. The data will display all companies that the companies in the universe made bids to.

    attribute : str, default 'companyname'
        Desired security attribute identifier to be shown for the counterparty company.

    currency : str, {'report', 'trade', ISO3 currency}, default 'report'
        | Desired currency for the financial data.

        - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
        - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
        - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        .. admonition:: Warning
            :class: warning

            | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
            | It will behave like parameter input currency=None

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> ma = prism.transaction.ma(710004, True, 'companyname')
        >>> prism.get_data(ma, 'Korea_primary', '2020-01-01', '2021-01-01')
             listingid        date          value     status   attitude      approach                                   acquirer  currency
        0     20120758  2020-12-17   22177.423520       Open   Friendly   Unsolicited                                        SBW       KRW
        1     20159169  2020-07-10   42698.255104       Paid   Friendly   Unsolicited  GF Financial Industry's No. 1 Corporation       KRW
        2     20159448  2020-03-25   19700.000000       Paid   Friendly   Unsolicited                     Magic Holdings Co.,Ltd       KRW
        3     20166644  2020-01-23   34350.877000  Withdrawn   Friendly   Unsolicited                        GSC. Korea Co.,Ltd.       KRW
        4     20177780  2020-04-13    5846.000000       Paid   Friendly   Unsolicited                      N&F Invesment Union 1       KRW
        ...        ...         ...            ...        ...        ...           ...                                        ...       ...
        210  632312955  2020-01-22  218108.748750       Paid   Friendly     Solicited                           Wintec Co., Ltd.       KRW
        211  636449999  2020-07-08  796474.922456       Paid   Friendly   Unsolicited            Skylake Equity Partners Limited       KRW
        212  643790044  2020-09-18   13792.675000  Withdrawn   Friendly     Solicited                              DAVOLINK Inc.       KRW
        213  671671272  2020-10-26  686100.000000  Withdrawn   Friendly   Unsolicited               Soulbrain Holdings Co., Ltd.       KRW
        214  671671272  2020-12-17  497756.173300       Paid   Friendly   Unsolicited               Soulbrain Holdings Co., Ltd.       KRW
    """

    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        target: bool,
        attribute: _CompanyRelAttributeType,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search : str = None, package : str = None):
        return cls._dataitems(search=search, package=package)


class buyback(_PrismTransactionComponent):
    """
    | Get buyback transaction-related data for equity securities.
    | The default frequency is aperiodic.

    Parameters
    ----------
    currency : str, {'report', 'trade', ISO3 currency}, default 'report'
        | Desired currency for the financial data.

        - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
        - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
        - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        .. admonition:: Warning
            :class: warning

            | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
            | It will behave like parameter input currency=None

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> buyback = prism.transaction.buyback()
        >>> prism.get_data(buyback, 'Korea_primary', '2020-01-01', '2021-01-01')
             listingid        date    value  currency
        0     20126254  2020-03-16   5000.0       KRW
        1     20130724  2020-03-20  40000.0       KRW
        2     20154597  2020-03-16   3000.0       KRW
        3     20158476  2020-06-18   4000.0       KRW
        4     20158750  2020-03-23    800.0       KRW
        ...        ...         ...      ...       ...
        232  644427262  2020-03-13   1000.0       KRW
        233  652274726  2020-03-30  10000.0       KRW
        234  652868489  2020-03-17   1000.0       KRW
        235  672007118  2020-11-03   1000.0       KRW
        236  675972663  2020-08-14   3000.0       KRW
    """

    @_validate_args
    def __init__(
        self,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search : str = None, package : str = None):
        return cls._dataitems(search=search, package=package)


@_validate_args
def dataitems(search: str = None, package: str = None):
    return _list_dataitem(
            datacategoryid=_PrismTransactionComponent.categoryid,
            datacomponentid=None,
            search=search,
            package=package,
        )


import pandas as pd

from .._req_builder import _list_dataitem, _dataquery
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils import _validate_args, _get_params, _req_call

__all__ = [
    "share",
    "weight",
    "level",
    "universe_dataitems",
    "portfolio_dataitems",
    "dataitems",
]


_data_category = __name__.split(".")[-1]


class _PrismIndexComponent(_PrismDataComponent, _PrismComponent):
    _component_category_repr = _data_category

    @classmethod
    def _dataitems(cls, search: str = None, package: str = None):
        ret = _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )
        ret = ret.drop(["dataitemdescription"], axis=1, errors="ignore")
        return ret

    @_validate_args
    @_req_call(_dataquery)
    def get_data(self, startdate: str = None, enddate: str = None, shownid = None, name = None,) -> pd.DataFrame:
        pass


class share(_PrismIndexComponent):
    """
    | Index constituent share data.
    | Default frequency is business daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.index.share.dataitems("Russell 3000 Index")
        >>> di[["dataitemid", "dataitemname"]]
           dataitemid        dataitemname
        0     4000099  Russell 3000 Index

        >>> rus = prism.index.share(4000099)
        >>> rus_df = rus.get_data(startdate='2010-01-01', enddate='2020-12-31', shownid=["Company Name"])
        >>> rus_df
                 listingid        date	     value	         Company Name
        0          2598345  2012-06-25  15863000.0  CARRIAGE SERVICES INC
        1          2598345  2012-06-26  15863000.0  CARRIAGE SERVICES INC
        2          2598345  2012-06-27  15863000.0  CARRIAGE SERVICES INC
        3          2598345  2012-06-28  15863000.0  CARRIAGE SERVICES INC
        4          2598345  2012-06-29  15863000.0  CARRIAGE SERVICES INC
        ...            ...         ...         ...                    ...
        2810518  403068703  2014-06-30  16860000.0   TRI POINTE HOMES INC
        2810519  403068703  2014-07-01  16860000.0   TRI POINTE HOMES INC
        2810520  403068703  2014-07-02  16860000.0   TRI POINTE HOMES INC
        2810521  403068703  2014-07-03  16860000.0   TRI POINTE HOMES INC
        2810522  403068703  2014-07-07  16860000.0   TRI POINTE HOMES INC
    """
    @_validate_args
    def __init__(self, dataitemid: int):
        return super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the share data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.
            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns :
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*


        Examples
        --------
            >>> di = prism.index.share.dataitems("Russell 3000 Index")
            >>> di[["dataitemid", "dataitemname"]]
            dataitemid        dataitemname
            0     4000099  Russell 3000 Index
        """
        return cls._dataitems(search=search, package=package)


class weight(_PrismIndexComponent):
    """
    | Index constituent weight data.
    | Default frequency is business daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.index.weight.dataitems("Russell 3000 Index")
        >>> di[["dataitemid", "dataitemname"]]
           dataitemid        dataitemname
        0     4000099  Russell 3000 Index

        >>> rus = prism.index.weight(4000099)
        >>> rus_df = rus.get_data(startdate='2010-01-01', enddate='2020-12-31', shownid=["Company Name"])
        >>> rus_df
                 listingid        date     value           Company Name
        0          2598345  2012-06-25  0.000009  CARRIAGE SERVICES INC
        1          2598345  2012-06-26  0.000009  CARRIAGE SERVICES INC
        2          2598345  2012-06-27  0.000009  CARRIAGE SERVICES INC
        3          2598345  2012-06-28  0.000009  CARRIAGE SERVICES INC
        4          2598345  2012-06-29  0.000009  CARRIAGE SERVICES INC
        ...            ...         ...       ...                    ...
        5989430  692043613  2020-12-23  0.000013         MEDIAALPHA INC
        5989431  692043613  2020-12-24  0.000013         MEDIAALPHA INC
        5989432  692043613  2020-12-28  0.000012         MEDIAALPHA INC
        5989433  692043613  2020-12-29  0.000011	     MEDIAALPHA INC
        5989434  692043613  2020-12-30  0.000011	     MEDIAALPHA INC
    """
    @_validate_args
    def __init__(self, dataitemid: int):
        return super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the weight data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns :
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.index.weight.dataitems("Russell 3000 Index")
            >>> di[["dataitemid", "dataitemname"]]
            dataitemid        dataitemname
            0     4000099  Russell 3000 Index
        """
        return cls._dataitems(search=search, package=package)


class level(_PrismIndexComponent):
    """
    | Index level data.
    | Default frequency is business daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        leveltype : str, default None, {'Price Return', 'Total Return Gross'}
            | Default value None gives all leveltype.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.index.level.dataitems("Russell 3000 Index")
        >>> di[["dataitemid", "dataitemname"]]
           dataitemid        dataitemname
        0     4000099  Russell 3000 Index

        >>> rus = prism.index.level(4000099)
        >>> rus_df = rus.get_data(startdate='2010-01-01', enddate='2020-12-31')
        >>> rus_df
                     date           leveltype       value
        0      2010-01-04  Total Return Gross  2925.36135
        1      2010-01-05  Total Return Gross  2933.81744
        2      2010-01-06  Total Return Gross  2937.46528
        3      2010-01-07  Total Return Gross  2949.90064
        4      2010-01-08  Total Return Gross  2959.37327
        ...           ...                 ...         ...
        11067  2020-12-23    Total Return Net  3243.92347
        11068  2020-12-24    Total Return Net  3252.86186
        11069  2020-12-28    Total Return Net  3270.69819
        11070  2020-12-29    Total Return Net  3257.80733
        11071  2020-12-30    Total Return Net  3266.55282
    """
    @_validate_args
    def __init__(self, dataitemid: int):
        return super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the level data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns :
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.index.level.dataitems("Russell 3000 Index")
            >>> di[["dataitemid", "dataitemname"]]
            dataitemid        dataitemname
            0     4000099  Russell 3000 Index
        """
        return cls._dataitems(search=search, package=package)

    @_validate_args
    @_req_call(_dataquery)
    def get_data(self, startdate: str = None, enddate: str = None, name = None,) -> pd.DataFrame:
        pass


@_validate_args
def universe_dataitems(search: str = None, package: str = None):
    """
    Usable data items for the index data category, which can be used to create universe in prism.save_index_as_universe()

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

        package : str, default None
            | Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to cash flow statement data component.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*

    Examples
    --------
    >>> di = prism.index.universe_dataitems("Korea Stock Price 200 Index")
    >>> di[["dataitemid", "dataitemname"]]
       dataitemid                 dataitemname
    0     6000034  Korea Stock Price 200 Index
    1     6000034  Korea Stock Price 200 Index
    """
    ret = _list_dataitem(datacategoryid=_PrismIndexComponent.categoryid, datacomponentid=17, search=search, package=package)
    ret = ret.drop(["dataitemdescription"], axis=1, errors="ignore")
    return ret


@_validate_args
def portfolio_dataitems(search: str = None, package: str = None):
    """
    Usable data items for the index data category, which can be used to create portfolio in prism.save_index_as_portfolio().

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

        package : str, default None
            | Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to cash flow statement data component.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*

    Examples
    --------
        >>> di = prism.index.portfolio_dataitems("Russell 3000 Index")
        >>> di[["dataitemid", "dataitemname"]]
           dataitemid        dataitemname
        0     4000099  Russell 3000 Index
    """
    ret = _list_dataitem(datacategoryid=_PrismIndexComponent.categoryid, datacomponentid=16, search=search, package=package)
    ret = ret.drop(["dataitemdescription"], axis=1, errors="ignore")
    return ret


@_validate_args
def dataitems(search: str = None, package: str = None):
    """
    Usable data items for the index data category.

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

        package : str, default None
            | Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to cash flow statement data component.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*

    Examples
    --------
    >>> di = prism.index.dataitems("Korea Stock Price 200 Index")
    >>> di[["dataitemid", "dataitemname"]]
       dataitemid                 dataitemname
    0     6000034  Korea Stock Price 200 Index
    1     6000034  Korea Stock Price 200 Index
    """
    ret = _list_dataitem(
        datacategoryid=_PrismIndexComponent.categoryid,
        datacomponentid=None,
        search=search,
        package=package,
    )
    ret = ret.drop(["dataitemdescription"], axis=1, errors="ignore")
    return ret



import copy
from typing import Union
import warnings

from ..._common.config import *
from ..._common.config import URL_TASK, URL_UNIVERSES
from ..._common import const
from ..._core._req_builder._universe import parse_universe_to_universeid
from ..._core._req_builder._exportdata import should_overwrite_datafile
from ..._utils import (
    get as _get,
    _validate_args,
    are_periods_exclusive,
    get_sm_attributevalue,
    Loader, post as _post, are_periods_exclusive,
)
from ..._utils.exceptions import PrismTaskError, PrismValueError, PrismTypeError, PrismAuthError
from ..._prismcomponent import abstract_prismcomponent
from ..._prismcomponent.prismcomponent import _PrismTaskComponent


_data_category = __name__.split(".")[-1]


class PrismDataTaskComponent(_PrismTaskComponent):
    _component_category_repr = _data_category


class export_data(_PrismTaskComponent):
    """
    Returns export_data task component which enables users to quickly retrieve and save data of the specified components.

    Parameters
    ----------
        component : PrismComponent or list
            | PrismComponent which hold the logic to query data.

        universe:
            | Universe name (*str*) or universe id (*int*) used to query data.
            | Some components do not require universe information (eg. Exchange Rate), in which case to be left None.

        startdate : str, default None
            | Start date of the data. The data includes start date.
            | If None, the start date of the universe is used. The same applies when the input is earlier than the universe start date.

        enddate : str, default None
            | End date of the data. The data excludes end date.
            | If None, the end date of the universe is used. The same applies when the input is later than the universe end date.

        shownid : list, default None
            | List of Security Master attributes to display with the data.
            | See prism securitymaster list_attribute for full list of Security Master attributes.
            | If None, default attributes set in preferences is shown.
            | If empty list ([]), no attribute is shown.

        display_pit: bool, default True
            | If True, universe data is left-joined to show missing (NA) values.
            | If False, the NA values are dropped.

        name : str or list, default None
            | Column names of the data to display.
            | If one component is passed to the function, accepts either string or list with one element.
            | If multiple components is passed to the function, accepts list of string.
            | If None:

            - If data component is passed, the column name is implicitly decided following the name of the data component.
            - If function component is passed, the default column name is 'value'.

    Returns
    -------
        ExportData component: prism._ExportData
            Prism Export Data Task Component.

    Examples
    --------
        >>> close = prism.market.close()
        >>> open = prism.market.open()
        >>> ed = prism.export_data([close, open], "KRX_300", "2022-01-01")
        ==== export_data
            Query Structure
        >>> ed.run("filepath/close", ["close", "open"])
        export_data is added to worker queue!
        {'status': 'Pending',
        'message': 'export_data is added to worker queue!',
        'result': [{'resulttype': 'jobid', 'resultvalue': 465}]}
    """
    @_validate_args
    def __init__(
        self,
        component: Union[abstract_prismcomponent._AbstractPrismComponent, list],
        universe: Union[str, int] = None,
        startdate: str = None,
        enddate: str = None,
        shownid: list = None,
        display_pit: bool = True,
        name: Union[str, list] = None,
    ):
        query = []
        cmpts = set()
        if isinstance(name, list) & (name is not None):
            if any([not isinstance(n, str) for n in name]):
                raise PrismTypeError('Names shoud be string')

        def add_cmpts(o):
            cmpts = set()
            if o["component_type"] == "datacomponent":
                cmpts.add(o["componentid"])
            else:
                for c in o["children"]:
                    cmpts = cmpts | add_cmpts(c)
            return cmpts

        if not isinstance(component, list):
            component = [component]
        for o in component:
            if isinstance(o, abstract_prismcomponent._AbstractPrismComponent):
                query.append(o._query)
                cmpts = add_cmpts(o._query)
            else:
                raise PrismTypeError(f"Please provide Components into export_data")

        if all((~const.DataComponents[const.DataComponents["componentid"].isin(cmpts)]["need_universe"]).tolist()):
            universeid = None
        else:
            universeid, _ = parse_universe_to_universeid(universe)

        universe_info = _get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

        if universe_period_violated:
            raise PrismValueError(
                f'Query period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        default_shownid = True
        if (shownid is not None) and (len(shownid) == 0):
            shownid = None
            default_shownid = False
        if shownid is not None:
            shownid = [get_sm_attributevalue(a) for a in shownid]
        component_names = set([c._query["component_name"] for c in component])
        if const.FunctionComponents is None:
            raise PrismAuthError("Please Login First")
        aggregatecomponents = set(const.FunctionComponents[const.FunctionComponents["is_aggregate"]]["component_name_repr"].unique())
        if (len(component_names - aggregatecomponents) == 0) & (shownid is not None):
            warnings.warn(f"Shownid will be ignored for: {list(component_names & aggregatecomponents)}")

        super().__init__(
            dataqueries=query,
            universeid=int(universeid),
            startdate=startdate,
            enddate=enddate,
            shownid=shownid,
            default_shownid=default_shownid,
            datanames=name,
            display_pit=display_pit,
        )

    @_validate_args
    def run(
        self,
        exportdatapath: str,
        commponent_names: list = None,
        jobname: str = None,
        startdate: str = None,
        enddate: str = None,
    ):
        """
        Enables users to quickly retrieve and save specified components of data.

        Parameters
        ----------
            exportdatapath: str
                | File path of the exported data.

            component_names: list
                | List of component names in Export Data Task Component
                | Names have to be the same order as data component list in Task Component.

            jobname : str
                | Name of the job when the task component is run.
                | If None, the default job name sets to screen_{jobid}.

            startdate : str, default None
                | Startdate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite startdate parameter in the task component.

            enddate : str, default None
                | Enddate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite enddate parameter in the task component.

        Returns
        -------
            status : dict
                | Returns 'Pending' status.
                | Screening task is added to system task queue.

        Examples
        --------
            >>> close = prism.market.close()
            >>> open = prism.market.open()
            >>> ed = prism.export_data([close, open], "KRX_300", "2022-01-01")
            ==== export_data
                Query Structure
            >>> ed.run("filepath/close", ["close", "open"])
            export_data is added to worker queue!
            {'status': 'Pending',
            'message': 'export_data is added to worker queue!',
            'result': [{'resulttype': 'jobid', 'resultvalue': 465}]}

        """
        should_overwrite, err_msg = should_overwrite_datafile(exportdatapath, "creating")
        if not should_overwrite:
            print(f"{err_msg}")
            return

        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        component_args.update({"universeid": int(universeid)})

        if any([not isinstance(n, str) for n in commponent_names]):
            raise PrismTypeError('Name for each component shoud be string')
        if len(commponent_names) != len(component_args["dataqueries"]):
            raise PrismValueError(
                f'Number of names must be equal to the number of components'
            )

        universe_info = _get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]
        component_args.update({"exportdatapath": exportdatapath + ".ped"})
        component_args.update({"cmpts": commponent_names})

        if startdate is not None:
            component_args["startdate"] = startdate
        if enddate is not None:
            component_args["enddate"] = enddate

        universe_period_violated = are_periods_exclusive(
            universe_startdate, universe_enddate, component_args.get("startdate"), component_args.get("enddate")
        )

        if universe_period_violated:
            raise PrismValueError(
                f'Query period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        query = {
            "component_type": self._query["component_type"],
            "component_name": self._query["component_name"],
            "component_args": component_args,
        }

        rescontent = None
        with Loader("Export Data Running... ") as l:
            try:
                rescontent = _post(URL_TASK + "/export_data", params={"jobname": jobname}, body=query)
            except:
                l.stop()
                raise PrismTaskError("Export Data has failed.")
            if rescontent["status"] != "Pending":
                l.stop()
                raise PrismTaskError("Export Data has failed.")

        print(f'{rescontent["message"]}')
        return rescontent

    @classmethod
    def list_job(cls):
        """
        List all export_data jobs.
        """
        return cls._list_job()


import os
import sys
import inspect
import IPython
from ._core import *
from ._core import _data, ols, _model
from ._core._req_builder import preference
from ._utils.version import __version__
from ._utils.validate_utils import handle_sys_exception, handle_jupyter_exception
from ._prismcomponent.abstract_prismcomponent import _AbstractPrismComponent
from ._prismcomponent.prismcomponent import _PrismComponent, _PrismFinancialComponent, _PrismDataComponent


_username = ''
quiet_mode = False

IPython.core.interactiveshell.InteractiveShell.showtraceback = handle_jupyter_exception(IPython.core.interactiveshell.InteractiveShell.showtraceback)
sys.excepthook = handle_sys_exception

_data_pkg_list = inspect.getmembers(_data, inspect.ismodule)
_model_list = inspect.getmembers(_model, inspect.ismodule)
_data_pkg_list = _data_pkg_list + _model_list
for pkg_name, pkg in _data_pkg_list:
    new_pkg_name = f"{__name__}.{pkg_name}"
    if new_pkg_name not in sys.modules.keys():
        pkg.__name__ = new_pkg_name  # Will have no effect; see below
        sys.modules[new_pkg_name] = pkg

# ols module
new_pkg_name = f"{__name__}.ols"
ols.__name__ = new_pkg_name
sys.modules[new_pkg_name] = ols

new_pkg_name = f"{__name__}.preference"
preference.__name__ = new_pkg_name
sys.modules[new_pkg_name] = preference



from enum import EnumMeta, Enum

class PrismError(Exception):
    """Exception raised for errors in module prism"""

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None

        super().__init__(self.message)

    def __str__(self):
        exception_type = "\033[91mPrismError\033[0m"
        return f"{exception_type}: {self.message}"


class PrismResponseError(Exception):
    """Exception raised for response errors from Prism BackEnd"""

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None

        super().__init__(self.message)

    def __str__(self):
        exception_type = "\033[91mPrismResponseError\033[0m"
        return f"{exception_type}: {self.message}"


class PrismTypeError(Exception):
    """Exception raised for type errors from type checker in validate_utils.py"""

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None

        super().__init__(self.message)

    def __str__(self):

        exception_type = "\033[91mPrismTypeError\033[0m"

        return f"{exception_type}: {self.message}"


class PrismValueError(Exception):
    """Exception raised for wrong Enum"""

    def __init__(self, *args, **kwargs):

        if args:
            self.message = args[0]
        else:
            self.message = None

        if kwargs:
            self.valid_list = kwargs.get("valid_list")
            self.invalids = kwargs.get("invalids")
        else:
            self.valid_list = []
            self.invalids = []

        super().__init__(self.message)

    def __str__(self):

        exception_type = "\033[91mPrismValueError\033[0m"
        if self.valid_list is not None and len(self.valid_list) > 0:
            extra_info_type = "\033[92mValid Arguments\033[0m"
            if isinstance(self.valid_list, (Enum, EnumMeta)):
                extra_info_message = [item.value for item in self.valid_list]
            elif isinstance(self.valid_list, list):
                extra_info_message = self.valid_list

            if self.invalids is not None:
                extra_info_message = list(set(extra_info_message) - set(self.invalids))

            return f"{exception_type}: {self.message}\n{extra_info_type} : {extra_info_message}"
        return f"{exception_type}: {self.message}"


class PrismAuthError(Exception):
    """ """

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None

        super().__init__(self.message)

    def __str__(self):

        exception_type = "\033[91mPrismAuthError\033[0m"

        return f"{exception_type}: {self.message}"


class PrismNotFoundError(Exception):
    """Exception raised for"""

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None
        super().__init__(self.message)

    def __str__(self):

        exception_type = "\033[91mPrismNotFoundError\033[0m"

        return f"{exception_type}: {self.message}"


class PrismTaskError(Exception):
    """Exception raised for"""

    def __init__(self, *args, **kwargs):
        if args:
            self.message = args[0]
        else:
            self.message = None
        super().__init__(self.message)

    def __str__(self):

        exception_type = "\033[91mPrismTaskError\033[0m"

        return f"{exception_type}: {self.message}"



import pytest


from tests.test_util.utils import get_secret
from sqlalchemy import create_engine


TEST_DB_NAME = "testcase"


@pytest.fixture(autouse=True)
def db_env():
    test_db = get_secret("testdb")
    test_db['dbname'] = TEST_DB_NAME
    url = 'postgresql://{}:{}@{}:{}/{}'.format(test_db['username'], test_db['password'], test_db['host'], test_db['port'], test_db['dbname'])
    engine = create_engine(url)

    yield engine

    engine.dispose()


@pytest.fixture(scope="session", autouse=True)
def prism():
    import prism
    prism.login("superuser", "L:3v[5a:mv8,z3Cf")
    return prism



import numbers

from prism._utils.exceptions import PrismTypeError as _PrismTypeError

from .._common.const import SPECIALVALUEMAP as _SPECIALVALUEMAP
from .._prismcomponent.prismcomponent import _PrismComponent, _PrismValue, _functioncomponent_builder
from .._prismcomponent.abstract_prismcomponent import _AbstractPrismComponent
from .._utils import _validate_args


__all__ = [
    "n_periods_beta",
    "n_periods_intercept",
    "n_periods_residual",
    "cross_sectional_beta",
    "cross_sectional_intercept",
    "cross_sectional_residual"
]


def _fn_builder(component_name, component_args, cmpts):
    cmpts_ = []
    for c in cmpts:
        if isinstance(c, _PrismComponent):
            cmpts_.append(c)
        elif isinstance(c, numbers.Real):
            other = _SPECIALVALUEMAP.get(other, other)
            cmpts_.append(_PrismValue(data=c))
        else:
            raise _PrismTypeError(f"Incompatible type for {component_name}: {type(c)}")

    return _functioncomponent_builder(component_name, component_args, *cmpts_)


@_validate_args
def n_periods_beta(x: _AbstractPrismComponent, y: _AbstractPrismComponent, n: int, min_periods: int = 1, fit_intercept: bool = True):
    """
    Beta coefficient in ordinary least squares simple linear regression.

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

        n : int
            Number of data points from the independent and dependent time series to include in the regression

        min_periods : int, default 1
            Minimum number of observations in window required to have a value; otherwise, result is ``np.nan``.

        fit_intercept : bool, default True
            Whether to calculate the intercept for the regression. If False, no intercept will be used in calculations (i.e. data is expected to be centered).

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.

    Examples
    --------
        >>> cash_flow = prism.financial.cash_flow(dataitemid=100513, periodtype='LTM', preliminary='ignore', currency='trade')
        >>> cash_flow_daily = close.resample("D", 365)
        >>> returns_daily = prism.market.close().n_periods_pct_change(1).resample("D", 365)
        >>> beta = prism.ols.n_periods_beta(cash_flow_daily, returns_daily, 252, 150)
        >>> beta.get_data(universe='Korea_primary', startdate='2010-01-01', enddate='2011-01-01')
                listingid	       date         value
        0         2666983  2010-01-01  6.596677e-09
        1         2666983  2010-01-02  6.573793e-09
        2         2666983  2010-01-03  6.551399e-09
        3         2666983  2010-01-04  6.769521e-09
        4         2666983  2010-01-05  6.746380e-09
        ...           ...         ...           ...
        619136  306481998  2010-12-28           NaN
        619137  306481998  2010-12-29           NaN
        619138  306481998  2010-12-30           NaN
        619139  306481998  2010-12-31           NaN
        619140  306481998  2011-01-01           NaN
    """
    return _fn_builder(
        "n_periods_beta",
        {"n": n, "min_periods": min_periods, "fit_intercept": fit_intercept},
        [x, y]
    )


@_validate_args
def n_periods_intercept(x: _AbstractPrismComponent, y: _AbstractPrismComponent, n: int, min_periods=1):
    """
    Residual for ordinary least squares simple linear regression

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

        n : int
            Number of data points from the independent and dependent time series to include in the regression

        min_periods : int, default 1
            Minimum number of observations in window required to have a value; otherwise, result is ``np.nan``.

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.

    Examples
    --------
        >>> cash_flow = prism.financial.cash_flow(dataitemid=100513, periodtype='LTM', preliminary='ignore', currency='trade')
        >>> cash_flow_daily = close.resample("D", 365)
        >>> returns_daily = prism.market.close().n_periods_pct_change(1).resample("D", 365)
        >>> intercept = prism.ols.n_periods_intercept(cash_flow_daily, returns_daily, 252, 150)
        >>> intercept.get_data(universe='Korea_primary', startdate='2010-01-01', enddate='2011-01-01')
                listingid	       date     value
        0         2666983  2010-01-01  0.016087
        1         2666983  2010-01-02  0.016010
        2         2666983  2010-01-03  0.015933
        3         2666983  2010-01-04  0.017095
        4         2666983  2010-01-05  0.016844
        ...           ...         ...       ...
        619136  306481998  2010-12-28       NaN
        619137  306481998  2010-12-29       NaN
        619138  306481998  2010-12-30       NaN
        619139  306481998  2010-12-31       NaN
        619140  306481998  2011-01-01       NaN
    """
    return _fn_builder(
        "n_periods_intercept",
        {"n": n, "min_periods": min_periods},
        [x, y]
    )


@_validate_args
def n_periods_residual(x: _AbstractPrismComponent, y: _AbstractPrismComponent, n: int, min_periods: int = 1, fit_intercept: bool = True):
    """
    Intercept in ordinary least squares simple linear regression.

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

        n : int
            Number of data points from the independent and dependent time series to include in the regression

        min_periods : int, default 1
            Minimum number of observations in window required to have a value; otherwise, result is ``np.nan``.

        fit_intercept : bool, default True
            Whether to calculate the intercept for the regression. If False, no intercept will be used in calculations (i.e. data is expected to be centered).

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.


    Examples
    --------
        >>> cash_flow = prism.financial.cash_flow(dataitemid=100513, periodtype='LTM', preliminary='ignore', currency='trade')
        >>> cash_flow_daily = close.resample("D", 365)
        >>> returns_daily = prism.market.close().n_periods_pct_change(1).resample("D", 365)
        >>> residual = prism.ols.n_periods_residual(cash_flow_daily, returns_daily, 252, 150)
        >>> residual.get_data(universe='Korea_primary', startdate='2010-01-01', enddate='2011-01-01')
                listingid	       date      value
        0         2666983  2010-01-01  -0.009541
        1         2666983  2010-01-02  -0.009664
        2         2666983  2010-01-03  -0.009787
        3         2666983  2010-01-04  -0.000018
        4         2666983  2010-01-05  -0.000230
        ...           ...         ...        ...
        619136  306481998  2010-12-28        NaN
        619137  306481998  2010-12-29        NaN
        619138  306481998  2010-12-30        NaN
        619139  306481998  2010-12-31        NaN
        619140  306481998  2011-01-01        NaN
    """
    return _fn_builder(
        "n_periods_residual",
        {"n": n, "min_periods": min_periods, "fit_intercept": fit_intercept},
        [x, y]
    )


@_validate_args
def cross_sectional_beta(x: _AbstractPrismComponent, y: _AbstractPrismComponent, fit_intercept: bool = True):
    """
    Beta coefficient in ordinary least squares simple linear regression.

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

        fit_intercept : bool, default True
            Whether to calculate the intercept for the regression. If False, no intercept will be used in calculations (i.e. data is expected to be centered).

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.

    """
    return _fn_builder(
        "cross_sectional_beta",
        {"fit_intercept": fit_intercept},
        [x, y]
    )


@_validate_args
def cross_sectional_intercept(x: _AbstractPrismComponent, y: _AbstractPrismComponent):
    """
    Intercept in ordinary least squares simple linear regression

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.

    """
    return _fn_builder(
        "cross_sectional_intercept",
        {},
        [x, y]
    )


@_validate_args
def cross_sectional_residual(x: _AbstractPrismComponent, y: _AbstractPrismComponent, fit_intercept: bool = True):
    """
    Residual for ordinary least squares simple linear regression.

    Parameters
    ----------
        x : PrismComponent
            Independent variable

        y : PrismComponent
            Dependent variable

        fit_intercept : bool, default True
            Whether to calculate the intercept for the regression. If False, no intercept will be used in calculations (i.e. data is expected to be centered).

    Returns
    -------
        prism._PrismComponent
            Resampled timeseries of the PrismComponent.
    """
    return _fn_builder(
        "cross_sectional_residual",
        {"fit_intercept": fit_intercept},
        [x, y]
    )


import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def _get_price(
    universe: str,
    adjustment: str | bool,
    offset: int = 0,
    package: str = 'Prism Market',
    engine=None,
):
    """
    valid adjustment =
        'all', 'split', 'dividend', True, False
    """
    return True


def test_adjustment_param():
    assert _get_price('S&P 500', "all", offset=0, package='Prism Market')

    # valid adjustment param
    assert _get_price('S&P 500', "all")
    assert _get_price('S&P 500', "split")
    assert _get_price('S&P 500', "dividend")
    assert _get_price('S&P 500', True)
    assert _get_price('S&P 500', False)

    assert _get_price('S&P 500', adjustment="all")
    assert _get_price('S&P 500', adjustment="split")
    assert _get_price('S&P 500', adjustment="dividend")
    assert _get_price('S&P 500', adjustment=True)
    assert _get_price('S&P 500', adjustment=False)


def test_wrong_adjustment_param():
    pass
    # wrong value
    # with pytest.raises(PrismValueError):
    #     assert _get_price('S&P 500', 'alll')

    # with pytest.raises(PrismValueError):
    #     assert _get_price('S&P 500', 'SPLIT')

    # with pytest.raises(PrismValueError):
    #     assert _get_price('S&P 500', 'true')

    # # wrong type
    # with pytest.raises(PrismTypeError):
    #     assert _get_price('S&P 500', adjustment=0)

    # with pytest.raises(PrismTypeError):
    #     assert _get_price('S&P 500', True, offset='zero')

    # with pytest.raises(PrismTypeError):
    #     assert _get_price('S&P 500', False, '0')



from typing import Union

import pytest

from prism._common.const import FrequencyType, UniverseFrequencyType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def get_universe(universe,
                 frequency: str = None,
                 shownid: list = None) -> UniverseFrequencyType:
    """
    valid frequency =
        'D', 'W', 'MS', 'SMS', 'SM', 'M', 'Q', 'QS', 'AS', 'A'
    """
    return frequency


def test_get_universe_frequency_param(mocker):
    mocker.patch("prism._common.const.SMValues", {
                 "apple": "prism_code1", "AAPL": "prism_code1", "prism39": "prism_code3"})

    assert get_universe("universe", frequency=None) is None

    # valid frequecy param
    assert get_universe("universe", frequency="D",
                        shownid=["AAPL"]) == UniverseFrequencyType.CALENDAR_DAY
    assert get_universe(
        "universe", frequency="W") == UniverseFrequencyType.WEEKS
    assert get_universe(
        "universe", frequency="MS") == UniverseFrequencyType.MONTH_START
    assert get_universe(
        "universe", frequency="SMS") == UniverseFrequencyType.SEMI_MONTH_START
    assert get_universe(
        "universe", frequency="SM") == UniverseFrequencyType.SEMI_MONTH_END
    assert get_universe(
        "universe", frequency="M") == UniverseFrequencyType.MONTH_END
    assert get_universe(
        "universe", frequency="Q") == UniverseFrequencyType.QUARTER_END
    assert get_universe(
        "universe", frequency="QS") == UniverseFrequencyType.QUARTER_START
    assert get_universe(
        "universe", frequency="AS") == UniverseFrequencyType.YEAR_START
    assert get_universe(
        "universe", frequency="A") == UniverseFrequencyType.YEAR_END


def test_get_universe_wrong_param(mocker):
    mocker.patch("prism._common.const.SMValues", {
                 "apple": "prism_code1", "AAPL": "prism_code1", "prism39": "prism_code3"})

    # wrong value
    with pytest.raises(Exception):
        assert get_universe(
            "universe", frequency="a") == UniverseFrequencyType.YEAR_END

    with pytest.raises(Exception):
        assert get_universe(
            "S&P 500", frequency="ms") == UniverseFrequencyType.MONTH_START

    with pytest.raises(Exception):
        assert get_universe("S&P 500", shownid=[
                            "AAPL"], frequency="random") is None

    # wrong type
    with pytest.raises(Exception):
        assert get_universe("S&P 500", frequency=222)

    with pytest.raises(Exception):
        assert get_universe("S&P 500", shownid=123)


@_validate_args
def factor_backtest(self,
                    jobname: Union[str, list] = None,
                    report: bool = True,
                    frequency: str = None) -> FrequencyType:
    """
    valid frequency =
        'N', 'U', 'us', 'L', 'ms', 'S', 'T', 'min', 'H', 'BH', 'D', 'BD', 'W', 'MS', 'BMS', 'SMS', 'SM', 'BM', 'M', 'Q', 'QS', 'BQ', 'BQS', 'AS', 'A'
    """
    return frequency


def test_frequency_param():
    assert factor_backtest("job1", frequency=None) is None

    # check type
    # assert isinstance(factor_backtest("job", frequency="N"), FrequencyType)

    # valid frequecy param
    assert factor_backtest("job", frequency="N") == FrequencyType.NANOSECONDS
    assert factor_backtest("job", frequency="U") == FrequencyType.MICROSECONDS
    assert factor_backtest(
        "job", frequency="us") == FrequencyType.MICROSECONDS_ALIAS
    assert factor_backtest("job", frequency="L") == FrequencyType.MILISECONDS
    assert factor_backtest(
        "job", frequency="ms") == FrequencyType.MILISECONDS_ALIAS
    assert factor_backtest("job", frequency="S") == FrequencyType.SECONDS
    assert factor_backtest("job", frequency="T") == FrequencyType.MINUTES
    assert factor_backtest(
        "job", frequency="min") == FrequencyType.MINUTES_ALIAS
    assert factor_backtest("job", frequency="H") == FrequencyType.HOURS
    assert factor_backtest(
        "job", frequency="BH") == FrequencyType.BUSINESS_HOURS
    assert factor_backtest("job", frequency="D") == FrequencyType.CALENDAR_DAY
    assert factor_backtest("job", frequency="BD") == FrequencyType.BUSINESS_DAY
    assert factor_backtest("job", frequency="W") == FrequencyType.WEEKS
    assert factor_backtest("job", frequency="MS") == FrequencyType.MONTH_START
    assert factor_backtest(
        "job", frequency="BMS") == FrequencyType.BUSINESS_MONTH_START
    assert factor_backtest(
        "job", frequency="SMS") == FrequencyType.SEMI_MONTH_START
    assert factor_backtest(
        "job", frequency="SM") == FrequencyType.SEMI_MONTH_END
    assert factor_backtest(
        "job", frequency="BM") == FrequencyType.BUSINESS_MONTH_END
    assert factor_backtest("job", frequency="M") == FrequencyType.MONTH_END
    assert factor_backtest("job", frequency="Q") == FrequencyType.QUARTER_END
    assert factor_backtest(
        "job", frequency="QS") == FrequencyType.QUARTER_START
    assert factor_backtest(
        "job", frequency="BQ") == FrequencyType.BUSINESS_QUARTER_END
    assert factor_backtest(
        "job", frequency="BQS") == FrequencyType.BUSINESS_QUARTER_START
    assert factor_backtest("job", frequency="AS") == FrequencyType.YEAR_START
    assert factor_backtest("job", frequency="A") == FrequencyType.YEAR_END


def test_wrong_frequency_param():
    # wrong value
    # with pytest.raises(Exception):
    #     assert factor_backtest("job", frequency="random_frequency")

    with pytest.raises(Exception):
        assert factor_backtest("job", frequency="a") == FrequencyType.YEAR_END

    # wrong type
    with pytest.raises(Exception):
        assert factor_backtest("job", frequency=123)

    with pytest.raises(Exception):
        assert factor_backtest("job", report=333)



from typing import Union

import pytest

from prism._prismcomponent.abstract_prismcomponent import _AbstractPrismComponent
from prism._common.const import FillnaMethodType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def fillna(self, value: Union[str, float, int, _AbstractPrismComponent] = None, method: FillnaMethodType = None, n: int = None):
    return True


def test_fillna_with_wrong_value():
    with pytest.raises(PrismTypeError):
        assert fillna(None, value=["a", "b"])

    with pytest.raises(PrismTypeError):
        assert fillna(None, value={"key": "val"})


def test_fillna_with_method():
    assert fillna(None, method='backfill')
    assert fillna(None, method='bfill')
    assert fillna(None, method='pad')
    assert fillna(None, method='ffill')


def test_fillna_wrong_method():
    with pytest.raises(PrismValueError):
        assert fillna(None, method='unknwon method')


def test_fillna_wrong_signature():
    with pytest.raises(PrismValueError):
        assert fillna(None, method=None)

    # "value" and "method" arguement coexists
    with pytest.raises(PrismValueError):
        assert fillna(None, value=1, method='pad')

    with pytest.raises(PrismValueError):
        assert fillna(None, value=1, method='random_method')



import pytest

from prism._common.const import EstimatePeriodType as _PeriodType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def consensus(
    dataitemid: int,
    periodtype: _PeriodType,
    periodout: int = 0,
    currency: str = 'trade',
) -> _PeriodType:
    """
    valid periodtype =
        'Annual', 'A', 'Semi-Annual', 'SA', 'Quarterly', 'Q', 'YTD', 'LTM', 'Non-Periodic', 'NTM', 'Q-SA'
    """
    return periodtype


def test_periodtype_param():
    pass
    # assert consensus(0, 'Annual', 0, 'trade') == _PeriodType.ANNUAL

    # # check type
    # assert isinstance(consensus(0, 'Annual', 0, 'trade'), _PeriodType)

    # # valid periodtype param
    # assert consensus(1, 'Annual') == _PeriodType.ANNUAL
    # assert consensus(2, 'A') == _PeriodType.A
    # assert consensus(3, 'Semi-Annual') == _PeriodType.SEMI_ANNUAL
    # assert consensus(4, 'SA') == _PeriodType.SA
    # assert consensus(5, 'Quarterly') == _PeriodType.QUARTERLY
    # assert consensus(6, 'Q') == _PeriodType.Q
    # assert consensus(7, 'YTD') == _PeriodType.YTD
    # assert consensus(8, 'LTM') == _PeriodType.LTM
    # assert consensus(9, 'Non-Periodic') == _PeriodType.NON_PERIODIC
    # assert consensus(10, 'NTM') == _PeriodType.NTM
    # assert consensus(11, 'Q-SA') == _PeriodType.QSA

    # assert consensus(1, periodtype='Annual') == _PeriodType.ANNUAL
    # assert consensus(2, periodtype='A') == _PeriodType.A
    # assert consensus(3, periodtype='Semi-Annual') == _PeriodType.SEMI_ANNUAL
    # assert consensus(4, periodtype='SA') == _PeriodType.SA
    # assert consensus(5, periodtype='Quarterly') == _PeriodType.QUARTERLY
    # assert consensus(6, periodtype='Q') == _PeriodType.Q
    # assert consensus(7, periodtype='YTD') == _PeriodType.YTD
    # assert consensus(8, periodtype='LTM') == _PeriodType.LTM
    # assert consensus(9, periodtype='Non-Periodic') == _PeriodType.NON_PERIODIC
    # assert consensus(10, periodtype='NTM') == _PeriodType.NTM
    # assert consensus(11, periodtype='Q-SA') == _PeriodType.QSA


def test_wrong_periodtype_param():
    pass
    # wrong value
    # with pytest.raises(PrismValueError):
    #     assert consensus(1, 'annual') == _PeriodType.ANNUAL

    # with pytest.raises(PrismValueError):
    #     assert consensus(1, 'a') == _PeriodType.A

    # with pytest.raises(PrismValueError):
    #     assert consensus(1, 'typo period')

    # with pytest.raises(PrismValueError):
    #     assert consensus(1, 'q-SA') == _PeriodType.QSA

    # # wrong type
    # with pytest.raises(PrismTypeError):
    #     assert consensus(1, periodtype=0, currency='won')

    # with pytest.raises(PrismTypeError):
    #     assert consensus(2, 0)

    # with pytest.raises(PrismTypeError):
    #     assert consensus(3, 'Annual', '0') == _PeriodType.ANNUAL



import copy
from typing import Union

from ..._common.config import URL_TASK, URL_UNIVERSES
from ..._common import const
from ..._core._req_builder._universe import should_overwrite_universe, parse_universe_to_universeid
from ..._prismcomponent import prismcomponent as pcmpt, abstract_prismcomponent
from ..._prismcomponent.prismcomponent import _PrismTaskComponent
from ..._utils import (
    get as _get,
    _validate_args,
    are_periods_exclusive as _are_periods_exclusive,
    Loader, post as _post,
)
from ..._utils.exceptions import PrismTaskError, PrismValueError, PrismTypeError


_data_category = __name__.split(".")[-1]


class screen(_PrismTaskComponent):

    _component_category_repr = _data_category

    @_validate_args
    def __init__(
        self,
        rule: abstract_prismcomponent._AbstractPrismComponent,
        universe: Union[int, str],
        frequency: str,
        startdate: str = None,
        enddate: str = None,
    ):
        if isinstance(rule, pcmpt._PrismDataComponent) or const.FunctionComponents[const.FunctionComponents["componentid"]==rule.componentid]["categoryname"].values[0] != "Logical":
            raise PrismTypeError("screen task only available to boolean operations.")
        universeid, _ = parse_universe_to_universeid(universe)

        universe_info = _get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        universe_period_violated = _are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

        if universe_period_violated:
            raise PrismValueError(
                f'Screen period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        super().__init__(
            rule_dataquery=[rule._query],
            universeid=int(universeid),
            frequency=frequency,
            startdate=startdate,
            enddate=enddate,
        )

    @_validate_args
    def run(
        self,
        newuniversename: str,
        jobname: str = None,
        frequency: str = None,
        startdate: str = None,
        enddate: str = None,
    ):
        """
        Enables users to quickly construct custom time-variant universes through user defined rules to evaluate over the specified startdate and endddate.

        Parameters
        ----------
            newuniversename : str
                Name of the universe to be created.

            jobname : str
                | Name of the job when the task component is run.
                | If None, the default job name sets to screen_{jobid}.

            frequency : str
                | Desired sampling frequency to resample.
                |
                | Frequency Format
                | Description: The 'frequency format' combines all the elements of the 'frequency' parameter, providing a comprehensive way to specify the resampling frequency.
                | Format: XBF-S
                |
                | Elements of the frequency Parameter
                | X: Resampling Interval (Optional, If not given, it is defaults to be 1, meaning resampling occurs every data point)
                | Description: The 'X' element represents the number at which resampling should be operated. It determines how often data should be resampled.
                |
                | B: Business Frequency (Optional, If not given, it is assumed to be a non-business frequency)
                | Description: The 'B' element indicates whether the resampling frequency should align with business days. If 'B' is included, it signifies a business frequency; otherwise, it's not considered.
                |
                | F: Frequency Type
                | Description: The 'F' element specifies the frequency type, which defines the unit of time for resampling. It can take on one of the following values: [D, W, M, Q, SA, A].
                |
                | S: Specific Part of Frequency (Optional)
                | Description: The 'S' element is an optional part of the 'frequency' parameter, and its presence depends on the chosen frequency type ('F'). It allows you to specify a particular aspect of the resampling frequency.
                |

                    | Specific Part for Each Frequency Type
                    | A (Annual) Frequency - MM/dd (Months and Day)
                    |    Example: A-12/15 (Resample every year on December 15th)
                    |
                    | M (Monthly), Q (Quarterly), SA (Semi-Annual) Frequencies - dd (Day of the Month)
                    |    Example: 3M-25 (Resample every 3 months on the 25th day of the month)
                    |
                    | W (Weekly) Frequency - Day of the Week (Case Insensitive)
                    |    Example: 2W-Fri (Resample every 2 weeks on Fridays)
                    |
                    | D (Daily) Frequency - N/A (If specific part is given it will raise an error)

                .. admonition:: Note
                    :class: note

                    | Result for certain frequencies
                    |
                    | For Q (Quarterly) frequency, the resampled results can only be for the months of March, June, September, and December.
                    | For SA (Semi-Annual) frequency, the resampled results can only be for the months of June and December.
                    | For dynamic months, use 3M instead of Q and 6M instead of SA.

                | Example
                - 3M-25: Resample every 3 months on the 25th day of the month.
                - 3D: Resample every 3 days.
                - 2BM-15: Resample every 2 months on the 15th day of the month, considering business days.
                - A-12/15: Resample every year on December 15th.
                - 2BQ-3: Resample every 2 quarters, on the 3rd day of the month, considering business days.

        lookback : int
            The periods to lookback are defined by the resampling frequency parameter. For example, if resampling to Monthly data, this will lookback *lookback* Months.

            .. admonition:: Note
                :class: note

                | When up-sampling, the lookback input parameter must be specified properly.
                |
                | For example, if resampling from Quarterly data to Monthly data, the lookback should be at least '3' or larger.
                | If set to only '1', then the lookback will not look far enough back to fill in every month and missing values will be left in the output time-series.
                | If no input is supplied, it will go back to the last data available and fill in every missing value in between.


            startdate : str, default None
                | Startdate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite startdate parameter in the task component.

            enddate : str, default None
                | Enddate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite enddate parameter in the task component.

        Returns
        -------
            status : dict
                | Returns 'Pending' status.
                | Screening task is added to system task queue.

        Examples
        --------
            >>> prism.list_universe()
            universeid                 universename  universetype   startdate     enddate
            0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
            1           2                      S&P 500         index  1700-01-01  2199-12-31
            2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
            3           4           Russell 3000 Index         index  1700-01-01  2199-12-31

            >>> mcap = prism.market.market_cap()
            >>> marketcap_rule = mcap.cross_sectional_rank() <= 200 # Top 200 market capitalization
            >>> snp_200_screen = prism.screen(
                    rule=marketcap_rule,
                    universename="S&P 500",
                    startdate="2010-01-01",
                    enddate="2015-01-01",
                    frequency="D",
                    )
            >>> snp_200_screen.run(newuniversename="snp_200")
            {'status': 'Pending',
            'message': 'screen pending',
            'result': [{'resulttype': 'jobid', 'resultvalue': 5}]}

            >>> prism.job_manager()
            >>> # Wait for the job 5 in GUI until its status changed to 'Completed'

            >>> prism.list_universe()
            universeid                 universename  universetype   startdate     enddate
            0           1  Korea Stock Price 200 Index         index  1700-01-01  2199-12-31
            1           2                      S&P 500         index  1700-01-01  2199-12-31
            2           3    Russell 3000 Growth Index         index  1700-01-01  2199-12-31
            3           4           Russell 3000 Index         index  1700-01-01  2199-12-31
            4           5                      snp_200         index  2010-01-01  2015-01-01
        """
        should_overwrite, err_msg = should_overwrite_universe(newuniversename, "screening")
        if not should_overwrite:
            print(f"{err_msg}")
            return
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        component_args.update({"universeid": int(universeid)})

        universe_info = _get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]
        component_args.update({"newuniversepath": newuniversename + ".puv"})

        if frequency is not None:
            component_args["frequency"] = frequency
        if startdate is not None:
            component_args["startdate"] = startdate
        if enddate is not None:
            component_args["enddate"] = enddate

        universe_period_violated = _are_periods_exclusive(
            universe_startdate, universe_enddate, component_args.get("startdate"), component_args.get("enddate")
        )

        if universe_period_violated:
            raise PrismValueError(
                f'Screen period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        query = {
            "component_type": self._query["component_type"],
            "componentid": self._query["componentid"],
            "component_args": component_args,
        }

        rescontent = None
        with Loader("Screen Running... ") as l:
            try:
                rescontent = _post(f"{URL_TASK}/{self.componentid}", params={"jobname": jobname}, body=query)
            except:
                l.stop()
                raise PrismTaskError("Screen has failed.")
            if rescontent["status"] != "Pending":
                l.stop()
                raise PrismTaskError("Screen has failed.")

        print(f'{rescontent["message"]}')
        return rescontent

    @classmethod
    def list_job(cls):
        """
        List all screen jobs.

        Returns
        -------
            pandas.DataFrame
                All screen jobs.
            Columns
                - *jobid*
                - *jobname*
                - *jobstatus*
                - *starttime*
                - *endtime*
                - *frequency*
                - *screeneduniverseid*
                - *screeneduniversepath*
                - *universepath*
                - *universeid*
                - *description*
                - *period*

        Examples
        --------
        >>> prism.screen_jobs()
        jobid  jobname  jobstatus                starttime                        endtime  frequency  screeneduniverseid  ...  universeid  description                   period
        0      1     None  Completed  2022-06-30 03:51:13.630  2022-06-30 03:51:18.388999936          Q                 9.0  ...         7.0         None  2010-01-01 ~ 2015-01-01
        1      2     None  Completed  2022-06-30 03:49:24.680  2022-06-30 03:49:28.556000000          Q                 8.0  ...	     7.0         None  2010-01-01 ~ 2015-01-01
        2      3     None  Completed  2022-06-27 17:37:27.386  2022-06-27 17:37:51.248000000          Q                 9.0  ...         7.0         None  2010-01-01 ~ 2015-01-01
        """
        return cls._list_job()



import os
import errno
import sys
import requests
import time
import uuid

from prism._utils.exceptions import PrismAuthError

from .._common.config import URL_WEB_AUTH, URL_DOC_AUTH, HEADERS, URL_REFRESH

def _authentication():
    global auth_token, refresh_token, access_token_time
    try:
        current_time = time.time()
        if globals().get('access_token_time') is None:
            raise PrismAuthError("Please Login First")
        if current_time - access_token_time > 1200:
            res = requests.post(url=URL_REFRESH, cookies={"refresh_token": refresh_token})
            _create_token(res)
        headers = HEADERS.copy()
        if 'auth_token' not in globals():
            raise PrismAuthError(f"Please Login First")
        token = auth_token
        headers.update({"Authorization": "Bearer {}".format(token)})
        req_id = str(uuid.uuid4())[:8]
        headers.update({"requestid": req_id})
    except FileNotFoundError:
        raise PrismAuthError(f"Please Login First")
    return headers


def _delete_token():
    global auth_token, refresh_token
    auth_token = ''
    refresh_token = ''


def _create_token(response: requests.models.Response):
    global auth_token, refresh_token, access_token_time
    token = response.json()["access_token"]
    ref_token = response.cookies.get("refresh_token")
    auth_token = token
    refresh_token = ref_token
    access_token_time = time.time()


def _find_file_path(file_name: str):
    sys.path.append(os.getcwd())
    for path in sys.path:
        if os.path.exists(path + "/" + file_name):
            return path + "/" + file_name
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_name)


class TokenDoesNotExistError(Exception):
    def __init__(self):
        super().__init__("You should first login to create a token")


def _get_web_authentication_token():
    headers = _authentication()
    res = requests.post(URL_WEB_AUTH, headers=headers)
    web_auth_token = res.json().get("access_token")
    return web_auth_token

def _get_document_authentication_token():
    headers = _authentication()
    res = requests.post(URL_DOC_AUTH, headers=headers)
    web_auth_token = res.json().get("access_token")
    return web_auth_token



import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def n_periods_intercept(x=None, y=None, n: int = 1, min_periods: int = 1):
    return True


def test_min_periods_param():
    assert n_periods_intercept(min_periods=1)

    # valid min_periods param
    assert n_periods_intercept(min_periods=1)
    assert n_periods_intercept(min_periods=2)
    assert n_periods_intercept(min_periods=9)
    assert n_periods_intercept(min_periods=9 // 3)
    assert n_periods_intercept(
        min_periods=9999999999999999999999999999999999999999999999999999999999999999999999999999999999)


def test_wrong_datetype_param():
    # wrong value
    with pytest.raises(PrismValueError):
        assert n_periods_intercept(min_periods=0)

    with pytest.raises(PrismValueError):
        assert n_periods_intercept(min_periods=-1)

    with pytest.raises(PrismValueError):
        assert n_periods_intercept(min_periods=20 * -1)

    # wrong type
    with pytest.raises(PrismTypeError):
        assert n_periods_intercept(min_periods=9 / 3)

    with pytest.raises(PrismTypeError):
        assert n_periods_intercept(min_periods="1")


def test_function_param_contains_min_periods():
    @_validate_args
    def sample_function(param_min_periods: int, min_periods_param: int = 1):
        return True

    assert sample_function(param_min_periods=1)
    assert sample_function(param_min_periods=2, min_periods_param=2)

    with pytest.raises(PrismValueError):
        assert sample_function(param_min_periods=-1)

    with pytest.raises(PrismValueError):
        assert sample_function(param_min_periods=1, min_periods_param=-1)



import requests
import hashlib
import uuid

import prism
from ..._common.config import *
from ..._utils import _validate_args, _create_token, get, _delete_token
from ..._common import const

__all__ = ['login']


@_validate_args
def login(username: str = '', password: str = '', credential_file: str = ''):
    """
    Log in to PrismStudio.

    Parameters
    ----------
        username : str, default ''
            A string representing the username of the user.
        password : str, default ''
            A string representing the password of the user.
        credential_file : str, default ''
            | Provide credential text files.
            | The first line of the file should be username and the second line should be password.
            | Please provide either valid credential file or username and password pair.

    Returns
    -------
        str
            A string with a success message if the login is successful, or **None** if the login fails.
    """
    if (not (username and password)) and (not credential_file):
        print("Please provide valid credential!")
        return
    if credential_file:
        with open(credential_file, 'r') as f:
            username = f.readline().strip()
            password = f.readline().strip()
    password = hashlib.sha512(password.encode())
    password = password.hexdigest()
    query = {'username': username, 'password': password}
    req_id = str(uuid.uuid4())[:8]
    headers = {"client": "python", 'requestid': req_id}
    res = requests.post(url=URL_LOGIN, data=query, headers=headers)

    if res.ok:
        _create_token(res)
        smattributes = get(f'{URL_SM}/attributes')
        const.SMValues = {a['attributerepr']: a['attribute'] for a in smattributes}
        const.PreferenceType = get(f'{URL_PREFERENCES}/types')
        const.CategoryComponent = get(URL_CATEGORYCOMPONENTS)
        const.FunctionComponents = get(URL_FUNCTIONCOMPONENTS)
        const.DataComponents = get(URL_DATACOMPONENTS)
        prism.username = username
        result = f'Login success! Welcome {username}'
    else:
        _delete_token()
        print(f'\033[91mLogin Failed\033[0m: Please check your username and password')
        return

    return result



from ..._common.config import *
from ..._common.const import *
from ..._utils import _validate_args, get


__all__ = ['dataitems', 'packages', 'functions']


@_validate_args
def dataitems(search : str = None, package : str = None):
    """
    Return all usable data items.

    Parameters
    ----------
        search : str, default None
            Search word for Data Items name, the search is case-insensitive.
        package : str, default None
            Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data Items that belong to the search results.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*

    Examples
    --------
        >>> import prism
        >>> prism.market.dataitems(search='short')
           dataitemid  ...                               dataitemname
        0     1100035  ...                Broker Short Interest Value
        1     1100055  ...        Short Interest Ratio (Day to Cover)
        2     1100056  ...                      Short Interest Tenure
        3     1100057  ...                       Short Interest Value
        4     1100058  ...          Short Interest as % Of Free Float
        5     1100059  ...  Short Interest as % Of Shares Outstanding
        6     1100060  ...                                Short Score
        7     1100063  ...           Supply Side Short Interest Value
    """

    param = {'search': search, 'package': package}
    return get(URL_DATAITEMS, param)


def packages():
    """
    Return all available packages.

    Returns
    -------
        pandas.DataFrame:
            Packages the user has access to.

        **Columns:**
            - packageid (*int*): Unique identifier for each package.
            - packagename (*str*): Name of each package.
            - source (*str*): Source of each package (e.g., CIQ, FactSet, or Prism39).
            - vendor (*str*): Name of the vendor that provides each package (e.g., S&P, FactSet, or Prism39).

    Examples
    --------
        >>> import prism
        >>> packages_df = prism.packages()
        >>> print(packages_df.head())
           packageid                packagename  source vendor
        0       1001   CIQ Alpha Factor Library     CIQ    S&P
        1       1002                   CIQ Base     CIQ    S&P
        2       1003  CIQ Business Relationship     CIQ    S&P
        3       1005   CIQ Company Intelligence     CIQ    S&P
        4       1006   CIQ Company Relationship     CIQ    S&P
    """
    return get(URL_PACKAGES, None)


def _list_dataitem(datacategoryid: str = None, datacomponentid: str = None, search : str = None, package : str = None):
    return get(
        URL_DATAITEMS,
        {'datacategoryid': datacategoryid, 'datacomponentid': datacomponentid, 'search': search, 'package': package}
    )



import pathlib
from enum import Enum
from pydantic import BaseSettings


class EnvironmentState(str, Enum):
    LOCAL = "local"
    DEVELOPMENT = "dev"
    STAGING = "stg"
    PRODUCTION = "production"
    DEMO = "demo"


class StateSettings(BaseSettings):
    ENV_STATE: str = 'production'
    VERSION: str = 'v1.3.0'
    URL_STABLE: str = 'https://api.prism39.com'
    URL_DEV: str = 'https://ops.prism39.com'
    dev: str = ':30495'
    stg: str = ':30496'
    demo: str = ':40452'

    class Config:
        p = pathlib.Path(__file__).parent.resolve()
        env_file = str(p) + '/.env'


class FactorySettings:
    @staticmethod
    def load():
        state = StateSettings()
        doc_url = 'help.prism39.com'
        ext_url = 'studio.prism39.com'
        if state.ENV_STATE == EnvironmentState.LOCAL:
            URL = "http://localhost:8000"
            doc_url = 'http://localhost:5500/doc/build'
            ext_url = "https://localhost:3000/ext"
        elif state.ENV_STATE == EnvironmentState.DEMO:
            URL = state.URL_STABLE + getattr(state, state.ENV_STATE)
            doc_url = f"https://{state.ENV_STATE}-help.prism39.com"
            ext_url = f"https://{state.ENV_STATE}-studio.prism39.com/ext"
        elif state.ENV_STATE in (EnvironmentState.DEVELOPMENT, EnvironmentState.STAGING):
            URL = state.URL_DEV + getattr(state, state.ENV_STATE)
            doc_url = f"https://{state.ENV_STATE}-help.prism39.com"
            ext_url = f"https://{state.ENV_STATE}-studio.prism39.com/ext"
        elif state.ENV_STATE == EnvironmentState.PRODUCTION:
            URL = state.URL_STABLE
            doc_url = "https://help.prism39.com"
            ext_url = "https://studio.prism39.com/ext"
        return URL, doc_url, ext_url


settings = FactorySettings.load()


__all__ = [
    'HEADERS',
    'URL_SERVICE',
    'URL_CATEGORYCOMPONENTS',
    'URL_DATACOMPONENTS',
    'URL_DATAQUERIES',
    'URL_TASKQUERIES',
    'URL_JOBS',
    'URL_PREFERENCES',
    'URL_PORTFOLIOS',
    'URL_UNIVERSES',
    'URL_DATAITEMS',
    'URL_FUNCTIONCOMPONENTS',
    'URL_PACKAGES',
    'URL_LOGIN',
    'URL_REFRESH',
    'URL_DBINFO',
    'URL_UPDATE',
    'URL_STATUS',
    'URL_TASK',
    'URL_VALIDATE',
    'URL_SM',
    'URL_WEB_AUTH',
    'URL_DOC_AUTH',
    'URL_STATIC',
    'URL_DATA_FILES',
    'ROOT_EXT_WEB_URL',
    'ROOT_DOCUMENT_URL',
    'URL_UPLOAD',
    "URL_IDS",
]

URL, ROOT_DOCUMENT_URL, ROOT_EXT_WEB_URL = settings

HEADERS = {'Content-Type': 'application/json; charset=utf-8', 'client': 'python'}

URL_SERVICE = URL + '/service'
URL_CATEGORYCOMPONENTS = URL + '/categorycomponents'
URL_DATAQUERIES = URL + '/dataqueries'
URL_DATACOMPONENTS = URL + '/datacomponents'
URL_TASKQUERIES = URL + '/taskqueries'
URL_JOBS = URL + '/jobs'
URL_PREFERENCES = URL + '/preference'
URL_UNIVERSES = URL + '/universes'
URL_PORTFOLIOS = URL + '/portfolios'
URL_IDS = URL + "/datacategorycomponentid"
URL_DATAITEMS = URL + '/dataitems'
URL_FUNCTIONCOMPONENTS = URL + '/functioncomponents'
URL_PACKAGES = URL + '/packages'
URL_LOGIN = URL + '/auth/login'
URL_REFRESH = URL + '/refresh_token'
URL_DBINFO = URL + '/dbinfo'
URL_UPDATE = URL + '/update'
URL_STATUS = URL + '/status'
URL_TASK = URL + '/task'
URL_VALIDATE = URL + '/validate'
URL_SM = URL + '/securitymaster'
URL_WEB_AUTH = URL + '/auth/web'
URL_DOC_AUTH = URL + '/auth/doc_web'
URL_STATIC = URL + '/static'
URL_UPLOAD = URL + '/upload'
URL_DATA_FILES = URL + '/datafiles'

_username = ''
# URL = 'https://api.prism39.com'



import pytest

from tests.test_util.utils import query_test


# -------------------------------------------------------------------------------------------------------------------- #
#                                                     Value Factors                                                    #
# -------------------------------------------------------------------------------------------------------------------- #
def test_sales_related_factors(sales_related_factors, db_env):
    query_test(sales_related_factors, db_env)


def test_book_related_factors(book_related_factors, db_env):
    query_test(book_related_factors, db_env)


def test_debt_related_factor(debt_related_factor, db_env):
    query_test(debt_related_factor, db_env)


def test_earnings_related_factors(earnings_related_factors, db_env):
    query_test(earnings_related_factors, db_env)


def test_cash_flow_related_factors(cash_flow_related_factors, db_env):
    query_test(cash_flow_related_factors, db_env)


def test_yield_related_factors(yield_related_factors, db_env):
    query_test(yield_related_factors, db_env)


# # -------------------------------------------------------------------------------------------------------------------- #
# #                                                     Growth Factors                                                    #
# # -------------------------------------------------------------------------------------------------------------------- #
def test_growth_sales_related_factor(growth_sales_related_factor, db_env):
    query_test(growth_sales_related_factor, db_env)


def test_growth_debt_related_factor(growth_debt_related_factor, db_env):
    query_test(growth_debt_related_factor, db_env)


def test_growth_earnings_related_factor(growth_earnings_related_factor, db_env):
    query_test(growth_earnings_related_factor, db_env)


def test_growth_cash_flow_related_factors(growth_cash_flow_related_factors, db_env):
    query_test(growth_cash_flow_related_factors, db_env)


def test_growth_yield_related_factors(growth_yield_related_factors, db_env):
    query_test(growth_yield_related_factors, db_env)


import json
import requests

import pandas as pd

from ..._common.config import *
from ..._common.const import *
from ..._utils import _validate_args, post, _authentication, _process_response


@_validate_args
def get_dbinfo(dbname: str):
    headers = _authentication()
    res = requests.get(url=URL_DBINFO + f'/{dbname}', headers=headers)
    try:
        res.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise SystemExit(json.loads(res.content)['detail'])
    result = pd.DataFrame(list(json.loads(res['data']).items()), columns=['name', 'value'])
    return result


@_validate_args
def delete_dbinfo(dbname: str):
    headers = _authentication()
    query = {'dbname': dbname}
    """need revision res.ok?"""
    res = requests.delete(url=URL_DBINFO + f'/{dbname}', params=json.dumps(query), headers=headers,)
    return _process_response(res)


@_validate_args
def create_dbinfo(
    dbname: str,
    host: str,
    port: str,
    db: str,
    dbusername: str,
    dbpassword: str,
    dialect: str,
    vendor: str,
    priority: int,
):
    body = {
        'dbname': dbname,
        'host': host,
        'port': port,
        'db': db,
        'dbusername': dbusername,
        'dbpassword': dbpassword,
        'dialect': dialect,
        'vendor': vendor,
        'priority': priority,
    }
    return post(URL_DBINFO, None, body)


@_validate_args
def update_dbinfo(
    dbname: str = None,
    host: str = None,
    port: str = None,
    db: str = None,
    dbusername: str = None,
    dbpassword: str = None,
    dialect: str = None,
    vendor: str = None,
    priority: int = None,
):
    headers = _authentication()
    query = {
        'dbname': dbname,
        'host': host,
        'port': port,
        'db': db,
        'dbusername': dbusername,
        'dbpassword': dbpassword,
        'dialect': dialect,
        'vendor': vendor,
        'priority': priority,
    }
    res = requests.patch(url=URL_DBINFO + f'/{dbname}', data=json.dumps(query), headers=headers)
    return _process_response(res)



from .._req_builder import _list_dataitem
from ..._common.const import CompanyRelAttributeType as _CompanyRelAttributeType
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils import _get_params, _validate_args
from ..._utils.exceptions import PrismValueError


__all__ = [
    "competitor",
    "description",
    "product",
    "relationship",
]


_data_category = __name__.split(".")[-1]


class _PrismCompanyComponent(_PrismDataComponent, _PrismComponent):
    _component_category_repr = _data_category


class competitor(_PrismCompanyComponent):
    """
    Retrieve data about equity securities' competitors.

    Parameters
    ----------
    target : bool
        Specifies whether the company is identified as a competitor by another company:
        - If True, the data will display all the companies that have identified the companies within the universe as competitors.
        - If False, the data will display all the companies that the companies within the universe have identified as competitors.

    attribute : str
        Desired security attribute identifier to be displayed for the counterparty company.

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> competitor = prism.company.competitor(True, 'companyname')
        >>> prism.get_data(competitor, 'Korea_primary', '2020-01-01', '2021-01-01')
              listingid                        value
        0      20108718                      BASF SE
        1      20108718            Adeka Corporation
        2      20108718             Chang Chun Group
        3      20108718          Sunko Ink Co., Ltd.
        4      20108718          Rianlon Corporation
        ...         ...                          ...
        4191  670598900              NEXON Co., Ltd.
        4192  670598900        Netmarble Corporation
        4193  693108646  Logitech International S.A.
        4194  693108646                    Cherry SE
        4195  693108646         Corsair Gaming, Inc.
    """
    @_validate_args
    def __init__(
        self,
        target: bool,
        attribute: _CompanyRelAttributeType,
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class description(_PrismCompanyComponent):
    """
    | Retrieve the business description of a company.

    Parameters
    ----------
    long : bool
        Specifies whether to display the extended version of the business description, if available.

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> desc = prism.company.description(True)
        >>> prism.get_data(desc, 'Korea_primary', '2020-01-01', '2021-01-01')
               listingid                                              value
        0       20108704  Daewon Pharmaceutical Co., Ltd., a pharmaceuti...
        1       20108706  N.I Steel Co., Ltd. produces and sells steel p...
        2       20108718  Songwon Industrial Co., Ltd., together with it...
        3       20108719  Whan In Pharm Co.,Ltd., a pharmaceutical compa...
        4       20109325  Theragen Etex Co.,Ltd., together with its subs...
        ...          ...                                                ...
        1850  1795845686  Saltware Co., Ltd. provides various IT service...
        1851  1801379557  Finger Story Co., Ltd. produces and distribute...
        1852  1819922381  Opticore.Inc manufactures and sells optical mo...
        1853  1824738416  Fine Circuit Co., Ltd. engages in manufacturin...
        1854  1830288132  Kostecsys. Co. Ltd. provides low thermal expan...
    """

    @_validate_args
    def __init__(
        self,
        long: bool,
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class product(_PrismCompanyComponent):
    """
    | Retrieve information about a company's available products.

    Parameters
    ----------

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> product = prism.company.product()
        >>> prism.get_data(product, 'Korea_primary', '2020-01-01', '2021-01-01')
                listingid        productname                                 productdescription  currentflag  updatedate
        0        20108704        Pelubi Tab.  Pelubi Tab an Anti-inflammatory, analgesic ind...            1         NaT
        1        20108704     Pelubi CR Tab.  Pelubi CR Tab. reduce symptoms of Osteoarthrit...            1         NaT
        2        20108704       Curefen Syr.  Curefen Syr. is an antipyretic analgesic, offe...            1         NaT
        3        20108704       Senafen Tab.  Senafen Tab. is indicated for Rheumatoid arthr...            1         NaT
        4        20108704  Tapain Tab./ Inj.  Tapain Tab./ Inj. is indicated for Rheumatoid ...            1         NaT
        ...           ...                ...                                                ...          ...         ...
        51031  1789113982       RF AQUASUNNY  Aquasunny RF system is an innovative injector ...            1         NaT
        51032  1789113982                FEI  Fei RF Needle offers a superior solution for s...            1         NaT
        51033  1789113982      CLABIANE THOR  CLABIANE THOR is Anti-aging microneedle skin r...            1         NaT
        51034  1789113982       Hifin Liften                                               None            1         NaT
        51035  1789113982          Hair Boom                                               None            1         NaT
    """
    @_validate_args
    def __init__(
        self,
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class relationship(_PrismCompanyComponent):
    """
    | Retrieve data about specific business relationships involving equity securities.

    Parameters
    ----------
    target : bool, default True
        Specifies whether the company is identified as having certain business relationships by other companies:
        - If True, the data will display all companies that have identified the companies within the universe with certain business relationships.
        - If False, the data will display all companies within the universe that have been identified with certain business relationships.

    attribute : str, default 'companyname'
        Desired security attribute identifier to be displayed for the counterparty company.

    Returns
    -------
    prism._PrismComponent

    Examples
    --------
        >>> rel = prism.company.relationship(True, 'companyname')
        >>> prism.get_data(rel, 'Korea_primary', '2020-01-01', '2021-01-01')
                listingid  businessreltypeid  sourcerelationship  targetrelationship  currentflag                                          source  updatedate
        0        20108704                  2            Supplier            Customer            0                               KX NexG Co., LTD.         NaT
        1        20108704                  4            Supplier         Distributor            0                                  Enzymotec Ltd.         NaT
        2        20108706                  2            Supplier            Customer            0  Suzhou Chunxing Precision Mechanical Co., Ltd.         NaT
        3        20108718                  2            Supplier            Customer            0                          SGC eTEC E&C Co., Ltd.         NaT
        4        20108718                  2            Supplier            Customer            0                           KOLON BENIT Co., Ltd.         NaT
        ...           ...                ...                 ...                 ...          ...                                             ...         ...
        19827   706837957                  8            Landlord              Tenant            0                         KB Securities Co., Ltd.         NaT
        19828   706837957                  2            Supplier            Customer            0                                        SSR Inc.         NaT
        19829   706837957                 20             Company  Strategic Alliance            0                                 HancomWITH Inc.         NaT
        19830  1675455809                  2            Supplier            Customer            0                                  METIS Co. Ltd.         NaT
        19831  1790081744                  4            Supplier         Distributor            0                        Eukor Car Carriers, Inc.         NaT
    """
    @_validate_args
    def __init__(
        self,
        target: bool,
        attribute: _CompanyRelAttributeType,
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))



import copy
import requests
from typing import Union

import prism
from prism._utils.exceptions import PrismNotFoundError, PrismResponseError, PrismValueError
from ..._common.config import *
from ..._common import const
from ..._core._req_builder import _dataquery
from .preference import get_quiet_mode
from ..._utils import _validate_args, post, patch, delete, _authentication, _fetch_and_parse, plot_tree


__all__ = [
    "list_taskquery",
    "load_taskquery",
    "save_taskquery",
    "delete_taskquery",
    "rename_taskquery",
    "extract_taskquery",
]


@_validate_args
def list_taskquery(search: str = None, tree=False):
    """
    Return the list of all saved task queries available.

    Parameters
    ----------
        search : str, default None
            | Search word for task query name, the search is case-insensitive.
            | If None, the entire list is returned.

        tree : bool, default False
            | If True, the folder structure of the data queries is visualized in a UNIX-style.

    Returns
    -------
        str or pandas.DataFrame
            | If tree is True, print file system tree of task queries.
            | If tree is False, return usable taskqueries in pandas.DataFrame.
            | Columns:

            - *taskqueryid*
            - *taskqueryname*

    Examples
    --------
        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	       screen
        1	           2	    ex_factor
        2	           3	  ex_strategy
    """

    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq", search)
    if taskqueries_df.empty:
        raise PrismNotFoundError("No Taskqueries found.")
    if not tree:
        return taskqueries_df
    plot_tree(taskqueries_df["taskqueryname"].tolist(), "taskquery")


@_validate_args
def load_taskquery(taskquery: Union[str, int]):
    """
    Load specified task queries.

    Parameters
    ----------
        taskquery : str or int
            Task query name (str) or task query id (int) to load.

    Returns
    -------
        prism._PrismComponent
            Return task query component.

    Examples
    --------
        >>> prism.load_taskquery(3)
        === Task Query Structure
    """

    headers = _authentication()
    taskqueryid = parse_taskquery_to_taskqueryid(taskquery)
    res = requests.get(url=URL_TASKQUERIES + f"/{taskqueryid}", headers=headers)
    if not res.ok:
        raise PrismResponseError(res.json()["message"])
    query = res.json()["rescontent"]["data"]["taskquerybody"]
    componentinfo = const.CategoryComponent[
        const.CategoryComponent["componentid"]==query["componentid"]
        & const.CategoryComponent["categoryid"]==query["categoryid"]
    ]
    component_name = componentinfo["component_name_repr"].values[0]
    component = getattr(prism, component_name)
    component_args = query["component_args"]
    for k in copy.deepcopy(list(component_args.keys())):
        if "_dataquery" in k:
            arg_key = k.split("_dataquery")[0]
            dataquery = _dataquery._dataquery_to_component(component_args.pop(k))
            component_args[arg_key] = dataquery
        if k.endswith("id"):
            component_args[k.split("id")[0]] = component_args.pop(k)
        if k == 'newuniversepath':
            component_args.pop('newuniversepath')
    result = component(**component_args)
    return result


@_validate_args
def save_taskquery(component, taskqueryname: str):
    """
    Save task query.

    Parameters
    ----------
        component : prism._PrismComponent
            New PrismComponent to be saved.

        taskqueryname : str
            Name of the new task query.

    Returns
    -------
        status: dict
            Status of save_taskquery.

    Examples
    --------
        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	       screen
        1	           2	    ex_factor
        2	           3	  ex_strategy

        >>> mcap = prism.market.market_cap()
        >>> marketcap_rule = mcap.cross_sectional_rank() <= 200 # Top 200 market capitalization
        >>> snp_200_screen = prism.screen(
                rule=marketcap_rule,
                universe="S&P 500",
                startdate="2010-01-01",
                enddate="2015-01-01",
                frequency="D",
                )
        >>> prism.save_taskquery(component=snp_200_screen,taskqueryname="mcap_200")
        {
        'status': 'Success',
        'message': 'Taskquery saved',
        'result': [{'resulttype': 'taskqueryid', 'resultvalue': 4}]
        }

        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	       screen
        1	           2	    ex_factor
        2	           3	  ex_strategy
        3	           4	     mcap_200
    """
    should_overwrite, err_msg = should_overwrite_taskquery(taskqueryname, "saving")
    if not should_overwrite:
        return err_msg
    ret = post(
        URL_TASKQUERIES,
        {
            "path": taskqueryname + ".ptq",
        },
        component._query,
    )
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_taskquery(old: Union[str, int], new: str):
    """
    Rename task query. Location of task query within the folder structure can be changed using the method.

    Parameters
    ----------
        old : str or int
            Name of existing task query (str) or task query id (int) to be renamed.

        new : str
            New name of the task query.

    Returns
    -------
        status: dict
            Status of rename_taskquery.

    Examples
    --------
        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	       screen
        1	           2	    ex_factor
        2	           3	  ex_strategy

        >>> prism.rename_taskquery(old="screen", name="new_screen")
        {
        'status': 'Success',
        'message': 'Taskquery renamed',
        'result': [{'resulttype': 'taskqueryid', 'resultvalue': 1}]
        }

        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	   new_screen
        1	           2	    ex_factor
        2	           3	  ex_strategy
    """
    should_overwrite, err_msg = should_overwrite_taskquery(new, "renaming")
    if not should_overwrite:
        return err_msg
    taskqueryid = parse_taskquery_to_taskqueryid(old)
    ret = patch(URL_TASKQUERIES + f"/{taskqueryid}", {"newpath": new + ".ptq"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_taskquery(taskquery: Union[str, int]):
    """
    Delete task query.


    Parameters
    ----------
        taskquery : str or int
            Task query name (str) or task query id (int) to delete.

    Returns
    -------
        status: dict
            Status of delete_taskquery.

    Examples
    --------
        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0  	         1	       screen
        1	           2	    ex_factor
        2	           3	  ex_strategy

        >>> prism.delete_taskquery(1)
        Taskquery deleted
        {
        'status': 'Success',
        'message': 'Taskquery deleted',
        'result': [{'resulttype': 'taskqueryid', 'resultvalue': 1}]
        }

        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0	           2	    ex_factor
        1	           3	  ex_strategy

        >>> prism.delete_taskquery("ex_factor")
        Taskquery deleted
        {
        'status': 'Success',
        'message': 'Taskquery deleted',
        'result': [{'resulttype': 'taskqueryid', 'resultvalue': 2}]
        }

        >>> prism.list_taskquery()
        taskqueryid  taskqueryname
        0	           3	  ex_strategy
    """
    dataqueryid = parse_taskquery_to_taskqueryid(taskquery)
    ret = delete(URL_TASKQUERIES + f"/{dataqueryid}")
    print(ret["message"])
    return ret


@_validate_args
def extract_taskquery(component, return_code=False):
    """
    Generate code which reproduces the task query provided by the component. If return_code is False, the method returns None and prints the code.

    Parameters
    ----------
        component : PrismComponent
            PrismComponent whose task query is to be extracted. Task component should be provided.

        return_code : bool, default False
            If True, the method returns the code. Else the code is printed as system output, returning None.

    Returns
    -------
        str or None
            | If return_code is True, return code in string.
            | If return_code is False, return None.

    Examples
    --------
        >>> mcap = prism.market.market_cap()
        >>> marketcap_rule = mcap.cross_sectional_rank() <= 200 # Top 200 market capitalization
        >>> snp_200_screen = prism.screen(
                rule=marketcap_rule,
                universe="S&P 500",
                startdate="2010-01-01",
                enddate="2015-01-01",
                frequency="D",
                )

        >>> prism.extract_dataquery(snp_200_screen)
        x0= prism.market.market_cap()
        x1 = x0.cross_sectional_rank() <= 200
        x2 = prism.screen(
                        rule=x1,
                universe="S&P 500",
                startdate="2010-01-01",
                enddate="2015-01-01",
                frequency="D",
                )
    """
    code = post(URL_TASKQUERIES + '/extract', {'dialect': 'python'}, component._query)
    if return_code:
        return code
    else:
        print(code)


@_validate_args
def parse_taskquery_to_taskqueryid(taskquery: Union[str, int]):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq")
    if taskqueries_df.empty:
        raise PrismNotFoundError("No Taskquery found.")

    if isinstance(taskquery, int):
        taskqueries_df = taskqueries_df[taskqueries_df["taskqueryid"] == taskquery]
    elif isinstance(taskquery, str):
        taskqueries_df = taskqueries_df[taskqueries_df["taskqueryname"] == taskquery]
    else:
        raise PrismValueError("Please provide taskquery path or taskquery id for taskquery.")

    if taskqueries_df.empty:
        raise PrismNotFoundError(f"No taskquery matching: {taskquery}")

    taskqueryid = taskqueries_df["taskqueryid"].values[0]
    return taskqueryid


def should_overwrite_taskquery(taskqueryname, operation):
    if get_quiet_mode():
        return True, None
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq")
    if taskqueries_df.empty:
        return True, None

    if taskqueryname in taskqueries_df["taskqueryname"].to_list():
        overwrite = input(f"{taskqueryname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Not {operation} taskquery."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None



from __future__ import annotations
from typing import Optional, List
from uuid import uuid4
import orjson
from pydantic import BaseModel, Field, UUID4, validator, root_validator


def to_camel(string: str) -> str:
    return ''.join(word.capitalize() if idx > 0 else word.lower() for idx, word in enumerate(string.split('_')))


def orjson_dumps(v, *, default):
    # orjson.dumps returns bytes, to match standard json.dumps we need to decode
    return orjson.dumps(v, default=default, option=orjson.OPT_NON_STR_KEYS).decode()


class PrismBaseModel(BaseModel):

    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True
        json_loads = orjson.loads
        json_dumps = orjson_dumps

    @root_validator(pre=True)
    def log_pre_validation(cls, values):
        return values


class PrismQuery(PrismBaseModel):
    nodeid: Optional[UUID4] = Field(default_factory=uuid4)
    component_type: str
    componentid: Optional[int]
    categoryid: Optional[int] = None
    component_name: Optional[str] = None
    component_category: Optional[str] = None
    component_args: Optional[dict] = {}
    children: Optional[List[PrismQuery]] = []




import pandas as pd

from ..._common.config import URL_PREFERENCES as _URL_PREFERENCES
from ..._common import const
from ._list import packages
from ..._utils import get as _get, patch as _patch, post as _post, _validate_args, get_sm_attributevalue
from ..._utils.exceptions import PrismTypeError, PrismValueError, PrismAuthError


__all__ = [
    "set_default_attributes",
    "get_default_attributes",
    "set_package_priority",
    "get_package_priority",
    "set_preference",
    "get_preference",
]


@_validate_args
def set_default_attributes(attributes: list):
    """
    Set default Security Master attributes to be used in data, universe, portfolio and reports.

    Parameters
    ----------
        attributes : list
            List of Security Master attributes to be set as default Security Master attributes.

    Returns
    -------
        status: dict
            Status of set_default_attributes.

    Examples
    --------
        >>> prism.preference.get_default_attributes()
        []

        >>> prism.preference.set_default_attributes(["CIQ Primary", "Country"])
        {
        'status': 'Success',
        'message': 'Attributes saved',
        'result': [{'resulttype': None, 'resultvalue': None}]
        }

        >>> prism.preference.get_default_attributes()
        ["CIQ Primary", "Country"]
    """
    attributevalues = [get_sm_attributevalue(a) for a in attributes]
    ret = _patch(_URL_PREFERENCES + "/attribute", {"attributes": attributevalues}, None)
    print(f'{ret["message"]}')
    print(f"Default attributes are: {get_default_attributes()}")
    return ret


@_validate_args
def get_default_attributes():
    """
    Return default Security Master attributes to show on data, universe, portfolio and reports.

    Returns
    -------
        list
            Current default Security Master attributes.

    Examples
    --------
        >>> prism.preference.get_default_attributes()
        ["Company Name", "ISIN", "Ticker"]
    """
    ga = _get(_URL_PREFERENCES + "/attribute")
    if const.SMValues is None:
        raise PrismAuthError(f"Please Login First")
    return [k for k, v in const.SMValues.items() if v in ga]


@_validate_args
def set_package_priority(datacategory: str, datacomponent: str, package_priority: list):
    """
    Set default package priorities to be used from each data component.

    Parameters
    ----------
        datacategory : str
            Data category to change the package priorities of. eg. financial, estimate, market.

        datacomponent : str
            Data component to change the package priorities of. eg. close, open, balance_sheet.

        package_priority : list of dict
            List of dictionary indicating the priorities of packages. Keys of the dictionary should be: “package” and “priority”.

    Returns
    -------
    status: dict
        Status of set_package_priority.

    Examples
    --------
        >>> datacategory = "Market"
        >>> datacomponent = "Close"
        >>> package_priority = [
            {"package": "CIQ Market", "priority": 1},
            {"package": "Prism Market", "priority": 2},
            {"package": "Compustat", "priority": 3},
            {"package": "MI Integrated Market", "priority": 4}
        ]
        >>> prism.preference.set_package_priority(datacategory=datacategory, datacomponent=datacomponent, package_priority=package_priority)
        {
        'status': 'Success',
        'message': 'Package priorities saved',
        'result': [{'resulttype': None, 'resultvalue': None}]
        }

        >>> priorities_df = prism.preference.get_package_priority()
        >>> priorities_df[priorities_df["datacomponent"] == "Close"]
            datacategory  datacomponent           packagename  priority
        0         Market          Close            CIQ Market         1
        1         Market          Close          Prism Market         1
        2         Market          Close             Compustat         1
        3         Market          Close  MI Integrated Market         1
    """

    if const.CategoryComponent is None:
        raise PrismAuthError(f"Please Login First")
    if datacomponent not in const.CategoryComponent[const.CategoryComponent["componenttype"]=="data"]["componentname"].unique():
        raise PrismValueError("Datacomponent should be one of prism data components.")
    old_priorities_df = get_package_priority()
    priorities = old_priorities_df[~((old_priorities_df['datacomponent'] == datacomponent) & (old_priorities_df['datacategory'] == datacategory))]
    new_priorities = []
    for i in package_priority:
        if not isinstance(i, dict):
            raise PrismTypeError("Elements of package priority should be a dictionary.")
        if {"package", "priority"} != set(i.keys()):
            raise PrismValueError("Package priority should have packagename and priority as keys.")
        new_priorities.append({"datacategory": datacategory, "datacomponent": datacomponent, "packagename": i["package"], "priority": i["priority"]})
    priorities = pd.concat([priorities, pd.DataFrame(new_priorities)])
    packages_df = packages()
    component_category = const.CategoryComponent.rename(columns={"componentname":"datacomponent", "categoryname": "datacategory"})
    priorities = priorities.merge(packages_df, on="packagename", how="left")
    priorities = priorities.merge(component_category, on=["datacategory", "datacomponent"], how="left")
    priorities = priorities[["componentid", "packageid", "priority"]]

    ret = _patch(_URL_PREFERENCES + "/package_priority", None, priorities.to_dict(orient='records'))
    print(f'{ret["message"]}')
    return ret


@_validate_args
def get_package_priority():
    """
    Return default package priorities to be used from each data component.

    Returns
    -------
        pandas.DataFrame
            | Package priorities to be used from each data component.
            | Columns:

            - *datacategory*
            - *datacomponent*
            - *packagename*
            - *priority*

    Examples
    --------
        >>> prism.preference.get_package_priority()
              datacategory         datacomponent               packagename  priority
        0         Estimate                Actual    CIQ Estimate Consensus         1
        1         Estimate             Consensus    CIQ Estimate Consensus         1
        2         Estimate                Growth    CIQ Premium Financials         1
        3         Estimate              Guidance    CIQ Estimate Consensus         1
        4         Estimate              Revision    CIQ Estimate Consensus         1
        ...            ...                   ...                     	 ...       ...
        71          Market                Volume                CIQ Market         4
        72          Market                Volume      MI Integrated Market         3
        73          Market                Volume                 Compustat         2
        74          Market                Volume              Prism Market         1
        75   Precalculated  Alpha Factor Library  CIQ Alpha Factor Library         1
    """
    return _get(_URL_PREFERENCES + "/package_priority")


@_validate_args
def set_preference(settings: dict):
    """
    Set default Security Master attributes to be used in data, universe, portfolio and reports.

    Parameters
    ----------
        settings : dict
            | Dictionary of settings to be set as preference.
            | Setting keys are: ['timezone', 'quiet_mode']

    Returns
    -------
        status : dict
            Status of set_preference.

    Examples
    --------
        >>> prism.preference.get_preference()
        preferencetype	 preferencevalue               updatedate
        0      quiet_mode	            true  2023-02-21 08:05:16.760
        1        timezone	             KST  2023-02-14 08:20:55.626

        >>> prism.preference.set_preference({ "quiet_mode": False, "timezone": "UTC" })
        {
        'status': 'Success',
        'message': 'Preference saved',
        'result': [{'resulttype': None, 'resultvalue': None}]
        }

        >>> prism.preference.get_preference()
           preferencetype	 preferencevalue               updatedate
        0      quiet_mode	           false  2023-02-21 08:05:16.760
        1        timezone	             UTC  2023-02-14 08:20:55.626
    """
    return _post(_URL_PREFERENCES + "/gui", params={}, body=settings)


@_validate_args
def get_preference(setting: str = ""):
    """
    Return preference of current user.

    Returns
    -------
        pandas.DataFrame
            | Setting of current user's preference.
            | Columns:

            - *preferencetype: preference keyword*
            - *preferencevalue: value of preference keyword*
            - *updatedate*

    Examples
    --------
        >>> prism.preference.get_preference()
           preferencetype	 preferencevalue               updatedate
        0      quiet_mode	           false  2023-02-21 08:05:16.760
        1        timezone	             UTC  2023-02-14 08:20:55.626

    """
    params = {}
    if setting:
        params = {"preferencetype": setting}
    return _get(_URL_PREFERENCES + "/gui", params=params)


def get_quiet_mode():
    quiet_mode = get_preference("quiet_mode")
    if quiet_mode.empty:
        return False
    return quiet_mode["preferencevalue"][0] == 'true'
